from pathlib import Path
import os

from parameterized import parameterized
import unittest

from .. import parse, dumps, UPLCDialect, eval
from ..util import NodeTransformer
from ..transformer import unique_variables
from ..optimizer import pre_evaluation

acceptance_test_path = Path("examples/acceptance_tests")


class AcceptanceTests(unittest.TestCase):
    # check that none of these transformers change semantics
    @parameterized.expand(
        [
            # No transformation -> Checks conformance of implementation of the CEK machine
            (NodeTransformer,),
            # Rewriting variable names for unique names
            (unique_variables.UniqueVariableTransformer,),
            # Pre-evaluating subterms - here it will always evaluate the whole expression as there are no missing variables
            (pre_evaluation.PreEvaluationOptimizer,),
        ]
    )
    def test_acceptance_tests(self, rewriter):
        for dirpath, dirs, files in sorted(os.walk(acceptance_test_path, topdown=True)):
            if dirs:
                # not a leaf directory
                continue
            with self.subTest("Acceptance test", path=dirpath):
                input_file = next(f for f in files if f.endswith("uplc"))
                input_file_path = os.path.join(dirpath, input_file)
                with open(input_file_path, "r") as fp:
                    input = fp.read()
                output_file = next(f for f in files if f.endswith("uplc.expected"))
                output_file_path = os.path.join(dirpath, output_file)
                with open(output_file_path, "r") as fp:
                    output = fp.read().strip()
                try:
                    input_parsed = parse(input, filename=input_file_path)
                except Exception:
                    self.assertEqual(
                        "parse error", output, "Parsing program failed unexpectedly"
                    )
                    continue
                try:
                    input_parsed = rewriter().visit(input_parsed)
                except unique_variables.FreeVariableError:
                    # will raise an evaluation error anyways
                    pass
                try:
                    res = eval(input_parsed)
                except Exception:
                    self.assertEqual(
                        "evaluation failure",
                        output,
                        "Evaluating program failed unexpectedly",
                    )
                    continue
                self.assertTrue(
                    output not in ("parse error", "evaluation failure"),
                    "Program parsed and evaluated but should have thrown error",
                )
                output_parsed = parse(output, filename=output_file_path).term
                res_parsed_unique = unique_variables.UniqueVariableTransformer().visit(
                    res
                )
                output_parsed_unique = (
                    unique_variables.UniqueVariableTransformer().visit(output_parsed)
                )
                res_dumps = dumps(res_parsed_unique, dialect=UPLCDialect.Aiken)
                output_dumps = dumps(output_parsed_unique, dialect=UPLCDialect.Aiken)
                self.assertEqual(
                    output_dumps, res_dumps, "Program evaluated to wrong output"
                )



from copy import copy

from ..util import NodeTransformer
from ..ast import *


class FreeVariableError(ValueError):
    pass


class UniqueVariableTransformer(NodeTransformer):
    def __init__(self):
        self.scopes = []
        self.count = 0

    def get_map(self, name: str):
        for orig, remap in reversed(self.scopes):
            if orig == name:
                return remap
        raise FreeVariableError(f"Variable {name} is never assigned")

    def push_map(self, name: str):
        new_name = f"v{self.count}"
        self.count += 1
        self.scopes.append((name, new_name))
        return new_name

    def pop_map(self):
        self.scopes.pop(-1)

    def visit_BoundStateLambda(self, node: BoundStateLambda):
        new_map = {}
        for k, v in node.state.items():
            nk = self.push_map(k)
            new_map[nk] = v
        n = self.push_map(node.var_name)
        nc = copy(node)
        nc.state = frozendict.frozendict(new_map)
        nc.var_name = n
        nc.term = self.visit(node.term)
        self.pop_map()
        for _ in node.state.keys():
            self.pop_map()
        return nc

    def visit_Lambda(self, node: Lambda):
        return self.visit_BoundStateLambda(node)

    def visit_BoundStateDelay(self, node: BoundStateDelay):
        for k in node.state.keys():
            self.push_map(k)
        nc = copy(node)
        nc.term = self.visit(node.term)
        for _ in node.state.keys():
            self.pop_map()
        return nc

    def visit_Variable(self, node: Variable):
        nc = copy(node)
        nc.name = self.get_map(node.name)
        return nc



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import logging


try:
    from .util import parse, eval, dumps, UPLCDialect

except ImportError as e:
    logging.error(
        "Error, trying to import dependencies. Should only occur upon package installation",
        exc_info=e,
    )

VERSION = (0, 5, 3)

__version__ = ".".join([str(i) for i in VERSION])
__author__ = "nielstron"
__author_email__ = "n.muendler@web.de"
__copyright__ = "Copyright (C) 2023 nielstron"
__license__ = "MIT"
__url__ = "https://github.com/imperatorlang/uplc"



import argparse
import enum
import json
import pathlib
import sys

import cbor2
import pyaiken
import pycardano

from .util import *
from .ast import Program, Apply
from .transformer import unique_variables


class Command(enum.Enum):
    eval = "eval"
    parse = "parse"
    build = "build"
    dump = "dump"


def main():
    a = argparse.ArgumentParser(
        description="An evaluator and compiler for UPLC written in python."
    )
    a.add_argument(
        "command",
        type=str,
        choices=Command.__members__.keys(),
        help="The command to execute on the input file.",
    )
    a.add_argument(
        "input_file",
        type=str,
        help="The input program to parse (foo.uplc). Set to - for stdin.",
    )
    a.add_argument(
        "-o",
        "--output-directory",
        default="",
        type=str,
        help="The output directory for artefacts of the build command. Defaults to the filename of the compiled contract. of the compiled contract.",
    )
    a.add_argument(
        "--dialect",
        default=UPLCDialect.Aiken.value,
        help="The dialect for dumping the parsed UPLC.",
        choices=[d.value for d in UPLCDialect],
    )
    a.add_argument(
        "--unique-varnames",
        action="store_true",
        help="Assign variables a unique name.",
    )
    a.add_argument(
        "--from-cbor",
        action="store_true",
        help="Read hex representation of flattened UPLC.",
    )
    a.add_argument(
        "args",
        nargs="*",
        default=[],
        help="Input parameters for the function, in case the command is eval.",
    )
    args = a.parse_args()
    command = Command(args.command)
    input_file = pathlib.Path(args.input_file) if args.input_file != "-" else sys.stdin
    with open(input_file, "r") as f:
        source_code = f.read()

    if args.from_cbor:
        source_code = pyaiken.uplc.unflat(source_code)
    code = parse(
        source_code,
        input_file.absolute() if isinstance(input_file, pathlib.Path) else None,
    )

    if command == Command.parse:
        print("Parsed successfully.")
        return

    if args.unique_varnames:
        code = unique_variables.UniqueVariableTransformer().visit(code)

    if command == Command.dump:
        print(dumps(code, UPLCDialect(args.dialect)))
        return

    if command == Command.build:
        if args.output_directory == "":
            if args.input_file == "-":
                print(
                    "Please supply an output directory if no input file is specified."
                )
                exit(-1)
            target_dir = pathlib.Path(pathlib.Path(input_file).stem)
        else:
            target_dir = pathlib.Path(args.output_directory)
        target_dir.mkdir(exist_ok=True)
        uplc_dump = code.dumps(dialect=UPLCDialect.Aiken)
        cbor_hex = pyaiken.uplc.flat(uplc_dump)
        # create cbor file for use with pycardano/lucid
        with (target_dir / "script.cbor").open("w") as fp:
            fp.write(cbor_hex)
        cbor = bytes.fromhex(cbor_hex)
        # double wrap
        cbor_wrapped = cbor2.dumps(cbor)
        cbor_wrapped_hex = cbor_wrapped.hex()
        # create plutus file
        d = {
            "type": "PlutusScriptV2",
            "description": f"",
            "cborHex": cbor_wrapped_hex,
        }
        with (target_dir / "script.plutus").open("w") as fp:
            json.dump(d, fp)
        script_hash = pycardano.plutus_script_hash(pycardano.PlutusV2Script(cbor))
        # generate policy ids
        with (target_dir / "script.policy_id").open("w") as fp:
            fp.write(script_hash.to_primitive().hex())
        addr_mainnet = pycardano.Address(
            script_hash, network=pycardano.Network.MAINNET
        ).encode()
        # generate addresses
        with (target_dir / "mainnet.addr").open("w") as fp:
            fp.write(addr_mainnet)
        addr_testnet = pycardano.Address(
            script_hash, network=pycardano.Network.TESTNET
        ).encode()
        with (target_dir / "testnet.addr").open("w") as fp:
            fp.write(addr_testnet)

        print(f"Wrote script artifacts to {target_dir}/")
        return
    if command == Command.eval:
        print("Starting execution")
        print("------------------")
        assert isinstance(
            code, Program
        ), "Main function must be wrapped in (program 1.0.0 ...)"
        try:
            f = code.term
            # UPLC lambdas may only take one argument at a time, so we evaluate by repeatedly applying
            for d in map(lambda a: parse(f"(program 1.0.0 {a})").term, args.args):
                f = Apply(f, d)
            ret = eval(f).dumps()
        except Exception as e:
            print("An exception was raised")
            ret = e
        print("------------------")
        print(ret)


if __name__ == "__main__":
    main()



from ast import iter_fields

import rply.errors

from .lexer import strip_comments, Lexer
from .parser import Parser
from .machine import Machine
from .ast import AST, UPLCDialect


def parse(s: str, filename=None):
    s = strip_comments(s)
    l = Lexer().get_lexer()
    p = Parser().get_parser()
    try:
        tks = l.lex(s)
        program = p.parse(tks)
    except rply.errors.LexingError as e:
        source = s.splitlines()[e.source_pos.lineno - 1]
        raise SyntaxError(
            f"Lexing failed, invalid token: {e.message}",
            (filename, e.source_pos.lineno, e.source_pos.colno, source),
        ) from None
    except rply.errors.ParsingError as e:
        source = s.splitlines()[e.source_pos.lineno - 1] if s else ""
        raise SyntaxError(
            f"Parsing failed, invalid production: {e.message}",
            (filename, e.source_pos.lineno, e.source_pos.colno, source),
        ) from None
    return program


def eval(u: AST):
    m = Machine(u)
    return m.eval()


def dumps(u: AST, dialect=UPLCDialect.Aiken):
    return u.dumps(dialect)


class NodeVisitor(object):
    """
    A node visitor base class that walks the abstract syntax tree and calls a
    visitor function for every node found.  This function may return a value
    which is forwarded by the `visit` method.

    This class is meant to be subclassed, with the subclass adding visitor
    methods.

    Per default the visitor functions for the nodes are ``'visit_'`` +
    class name of the node.  So a `TryFinally` node visit function would
    be `visit_TryFinally`.  This behavior can be changed by overriding
    the `visit` method.  If no visitor function exists for a node
    (return value `None`) the `generic_visit` visitor is used instead.

    Don't use the `NodeVisitor` if you want to apply changes to nodes during
    traversing.  For this a special visitor exists (`NodeTransformer`) that
    allows modifications.
    """

    def visit(self, node):
        """Visit a node."""
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        for field, value in iter_fields(node):
            self.visit(value)


class NodeTransformer(NodeVisitor):
    """
    A :class:`NodeVisitor` subclass that walks the abstract syntax tree and
    allows modification of nodes.

    The `NodeTransformer` will walk the AST and use the return value of the
    visitor methods to replace or remove the old node.  If the return value of
    the visitor method is ``None``, the node will be removed from its location,
    otherwise it is replaced with the return value.  The return value may be the
    original node in which case no replacement takes place.

    Here is an example transformer that rewrites all occurrences of name lookups
    (``foo``) to ``data['foo']``::

       class RewriteName(NodeTransformer):

           def visit_Name(self, node):
               return Subscript(
                   value=Name(id='data', ctx=Load()),
                   slice=Constant(value=node.id),
                   ctx=node.ctx
               )

    Keep in mind that if the node you're operating on has child nodes you must
    either transform the child nodes yourself or call the :meth:`generic_visit`
    method for the node first.

    For nodes that were part of a collection of statements (that applies to all
    statement nodes), the visitor may also return a list of nodes rather than
    just a single node.

    Usually you use the transformer like this::

       node = YourTransformer().visit(node)
    """

    def generic_visit(self, node):
        for field, old_value in iter_fields(node):
            new_node = self.visit(old_value)
            if new_node is None:
                delattr(node, field)
            else:
                setattr(node, field, new_node)
        return node



"""
DeBrujin Machine to evaluate UPLC AST
"""

import frozendict
import logging
from dataclasses import dataclass

from .ast import *

_LOGGER = logging.getLogger(__name__)


class Machine:
    def __init__(self, program: AST, max_steps=1000000):
        self.program = program
        self.rem_steps = max_steps

    def eval(self):
        stack = [
            Compute(
                NoFrame(),
                frozendict.frozendict(),
                self.program,
            )
        ]

        while stack:
            self.rem_steps -= 1
            if self.rem_steps < 0:
                raise RuntimeError("Maximum steps exceeded")
            step = stack.pop()
            if isinstance(step, Compute):
                stack.append(step.term.eval(step.ctx, step.env))
            elif isinstance(step, Return):
                stack.append(self.return_compute(step.context, step.value))
            elif isinstance(step, Done):
                stack.append(step.term)
                break

        return stack.pop()

    def return_compute(self, context, value):
        if isinstance(context, FrameApplyFun):
            return self.apply_evaluate(context.ctx, context.val, value)
        if isinstance(context, FrameApplyArg):
            return Compute(
                FrameApplyFun(
                    value,
                    context.ctx,
                ),
                context.env,
                context.term,
            )
        if isinstance(context, FrameForce):
            return self.force_evaluate(context.ctx, value)
        if isinstance(context, NoFrame):
            term = value
            return Done(term)

    def apply_evaluate(self, context, function, argument):
        if isinstance(function, BoundStateLambda):
            return Compute(
                context,
                function.state | {function.var_name: argument},
                function.term,
            )
        if isinstance(function, ForcedBuiltIn):
            eval_fun = BuiltInFunEvalMap[function.builtin]
            needs_forces = BuiltInFunForceMap[function.builtin]
            if function.applied_forces == needs_forces:
                if eval_fun.__code__.co_argcount == len(function.bound_arguments) + 1:
                    res = eval_fun(*function.bound_arguments, argument)
                else:
                    res = ForcedBuiltIn(
                        function.builtin,
                        function.applied_forces,
                        function.bound_arguments + [argument],
                    )
                return Return(context, res)
        raise RuntimeError("Tried to apply arguments to unapplyiable object")

    def force_evaluate(self, context, value):
        if isinstance(value, BoundStateDelay):
            return Compute(context, value.state, value.term)
        if isinstance(value, ForcedBuiltIn):
            needs_forces = BuiltInFunForceMap[value.builtin]
            if value.applied_forces < needs_forces:
                res = ForcedBuiltIn(
                    value.builtin, value.applied_forces + 1, value.bound_arguments
                )
                # Theoretically we could check if the builtin requires 0 arguments, but this is not the case for any function
                return Return(context, res)
        raise RuntimeError("Forcing an unforceable object")



import dataclasses
import enum
import json
import logging
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum, auto
import hashlib
from typing import List, Any, Dict, Union

import cbor2
import frozendict
import frozenlist
import nacl.exceptions
from pycardano.crypto.bip32 import BIP32ED25519PublicKey
from nacl import bindings


class UPLCDialect(enum.Enum):
    Aiken = "aiken"
    Plutus = "plutus"


class Context:
    pass


@dataclass
class FrameApplyFun(Context):
    val: Any
    ctx: Context


@dataclass
class FrameApplyArg(Context):
    env: frozendict.frozendict
    term: "AST"
    ctx: Context


@dataclass
class FrameForce(Context):
    ctx: Context


@dataclass
class NoFrame(Context):
    pass


class Step:
    pass


@dataclass
class Return:
    context: Context
    value: Any


@dataclass
class Compute:
    ctx: Context
    env: frozendict.frozendict
    term: "AST"


@dataclass
class Done:
    term: "AST"


_LOGGER = logging.getLogger(__name__)


class AST:
    _fields = []

    def eval(self, context: Context, state: frozendict.frozendict):
        raise NotImplementedError()

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        raise NotImplementedError()


@dataclass(frozen=True)
class Constant(AST):
    def eval(self, context, state):
        return Return(context, self)

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        return f"(con {self.typestring(dialect=dialect)} {self.valuestring(dialect=dialect)})"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        raise NotImplementedError()

    def typestring(self, dialect=UPLCDialect.Aiken):
        raise NotImplementedError()


@dataclass(frozen=True)
class BuiltinUnit(Constant):
    def typestring(self, dialect=UPLCDialect.Aiken):
        return "unit"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        return "()"


@dataclass(frozen=True)
class BuiltinBool(Constant):
    value: bool

    def typestring(self, dialect=UPLCDialect.Aiken):
        return "bool"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        return str(self.value)


@dataclass(frozen=True)
class BuiltinInteger(Constant):
    value: int

    def typestring(self, dialect=UPLCDialect.Aiken):
        return "integer"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        return str(self.value)

    def __add__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to add two non-builtin-integers"
        return BuiltinInteger(self.value + other.value)

    def __sub__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to sub two non-builtin-integers"
        return BuiltinInteger(self.value - other.value)

    def __mul__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to mul two non-builtin-integers"
        return BuiltinInteger(self.value * other.value)

    def __floordiv__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to floordiv two non-builtin-integers"
        return BuiltinInteger(self.value // other.value)

    def __mod__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to mod two non-builtin-integers"
        return BuiltinInteger(self.value % other.value)

    def __neg__(self):
        return BuiltinInteger(-self.value)

    def __eq__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to eq two non-builtin-integers"
        return BuiltinBool(self.value == other.value)

    def __le__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to le two non-builtin-integers"
        return BuiltinBool(self.value <= other.value)

    def __lt__(self, other):
        assert isinstance(
            other, BuiltinInteger
        ), "Trying to lt two non-builtin-integers"
        return BuiltinBool(self.value < other.value)


@dataclass(frozen=True)
class BuiltinByteString(Constant):
    value: bytes

    def typestring(self, dialect=UPLCDialect.Aiken):
        return "bytestring"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        return f"#{self.value.hex()}"

    def __add__(self, other):
        assert isinstance(
            other, BuiltinByteString
        ), "Trying to add two non-builtin-bytestrings"
        return BuiltinByteString(self.value + other.value)

    def __len__(self):
        return BuiltinInteger(len(self.value))

    def __eq__(self, other):
        assert isinstance(
            other, BuiltinByteString
        ), "Trying to eq two non-builtin-bytestrings"
        return BuiltinBool(self.value == other.value)

    def __le__(self, other):
        assert isinstance(
            other, BuiltinByteString
        ), "Trying to le two non-builtin-bytestrings"
        return BuiltinBool(self.value <= other.value)

    def __lt__(self, other):
        assert isinstance(
            other, BuiltinByteString
        ), "Trying to lt two non-builtin-bytestrings"
        return BuiltinBool(self.value < other.value)

    def __getitem__(self, item):
        if isinstance(item, BuiltinInteger):
            assert 0 <= item.value <= len(self.value), "Out of bounds"
            return BuiltinInteger(self.value[item.value])
        raise NotImplementedError()

    def decode(self, *args):
        return BuiltinString(self.value.decode("utf8"))


@dataclass(frozen=True)
class BuiltinString(Constant):
    value: str

    def typestring(self, dialect=UPLCDialect.Aiken):
        return "string"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        return json.dumps(self.value)

    def __add__(self, other):
        assert isinstance(other, BuiltinString), "Can only add two bytestrings"
        return BuiltinString(self.value + other.value)

    def __eq__(self, other):
        assert isinstance(
            other, BuiltinString
        ), "Can only compare two bytestrings for equality"
        return BuiltinBool(self.value == other.value)

    def encode(self, *args):
        return BuiltinByteString(self.value.encode())


@dataclass(frozen=True)
class BuiltinPair(Constant):
    l_value: Constant
    r_value: Constant

    def typestring(self, dialect=UPLCDialect.Aiken):
        if dialect == UPLCDialect.Aiken:
            return f"pair<{self.l_value.typestring(dialect=dialect)}, {self.r_value.typestring(dialect=dialect)}>"
        elif dialect == UPLCDialect.Plutus:
            return f"(pair {self.l_value.typestring(dialect=dialect)} {self.r_value.typestring(dialect=dialect)})"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        if dialect == UPLCDialect.Aiken:
            return f"[{self.l_value.valuestring(dialect=dialect)}, {self.r_value.valuestring(dialect=dialect)}]"
        elif dialect == UPLCDialect.Plutus:
            return f"({self.l_value.valuestring(dialect=dialect)}, {self.r_value.valuestring(dialect=dialect)})"

    def __getitem__(self, item):
        if isinstance(item, int):
            if item == 0:
                return self.l_value
            elif item == 1:
                return self.r_value
        raise ValueError()


@dataclass(frozen=True, init=False)
class BuiltinList(Constant):
    values: List[Constant]
    # dirty hack to handle the type of empty lists
    sample_value: Constant

    def __init__(self, values, sample_value=None):
        object.__setattr__(self, "values", values)
        if not values:
            assert (
                sample_value is not None
            ), "Need to provide a sample value for empty lists to infer the type"
            object.__setattr__(self, "sample_value", sample_value)
        else:
            object.__setattr__(self, "sample_value", values[0])

    def typestring(self, dialect=UPLCDialect.Aiken):
        if dialect == UPLCDialect.Aiken:
            return f"list<{self.sample_value.typestring(dialect=dialect)}>"
        elif dialect == UPLCDialect.Plutus:
            return f"(list {self.sample_value.typestring(dialect=dialect)})"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        return f"[{', '.join(v.valuestring(dialect=dialect) for v in self.values)}]"

    def __add__(self, other):
        assert isinstance(other, BuiltinList), "Can only append two lists"
        assert (
            other.typestring() == self.typestring()
        ), f"Expected {self.typestring()} but got {other.typestring()}"
        return BuiltinList(self.values + other.values)

    def __eq__(self, other):
        assert isinstance(other, BuiltinList), "Can only compare two lists"
        return self.values == other.values

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.values[item]
        elif isinstance(item, slice):
            return BuiltinList(self.values[item], self.sample_value)


@dataclass(frozen=True)
class PlutusData(Constant):
    pass

    def typestring(self, dialect=UPLCDialect.Aiken):
        return "data"

    def valuestring(self, dialect=UPLCDialect.Aiken):
        return f"#{cbor2.dumps(self.to_cbor()).hex()}"

    def to_cbor(self) -> bytes:
        """Returns a CBOR encodable representation of this object"""
        raise NotImplementedError


@dataclass(frozen=True)
class PlutusAtomic(PlutusData):
    value: Any

    def to_cbor(self):
        return self.value


@dataclass(frozen=True, eq=True)
class PlutusInteger(PlutusAtomic):
    value: int


@dataclass(frozen=True, eq=True)
class PlutusByteString(PlutusAtomic):
    value: bytes


@dataclass(frozen=True, eq=True)
class PlutusList(PlutusData):
    value: Union[List[PlutusData], frozenlist.FrozenList]

    def to_cbor(self):
        return [d.to_cbor() for d in self.value]


@dataclass(frozen=True, eq=True)
class PlutusMap(PlutusData):
    value: Union[Dict[PlutusData, PlutusData], frozendict.frozendict]

    def to_cbor(self):
        return {k.to_cbor(): v.to_cbor() for k, v in self.value.items()}


@dataclass(frozen=True, eq=True)
class PlutusConstr(PlutusData):
    constructor: int
    fields: Union[List[PlutusData], frozenlist.FrozenList]

    def to_cbor(self):
        if 0 <= self.constructor < 7:
            return cbor2.CBORTag(
                self.constructor + 121, [f.to_cbor() for f in self.fields]
            )
        elif 7 <= self.constructor < 128:
            return cbor2.CBORTag(
                (self.constructor - 7) + 1280, [f.to_cbor() for f in self.fields]
            )
        else:
            return cbor2.CBORTag(
                102, [self.constructor, [f.to_cbor() for f in self.fields]]
            )


def data_from_cbortag(cbor) -> PlutusData:
    if isinstance(cbor, cbor2.CBORTag):
        if 121 <= cbor.tag <= 121 + 6:
            constructor = cbor.tag - 121
            fields = cbor.value
        elif 1280 <= cbor.tag <= 1280 + (127 - 7):
            constructor = cbor.tag - 1280 + 7
            fields = cbor.value
        elif cbor.tag == 102:
            constructor, fields = cbor.value
        else:
            raise ValueError(f"Invalid cbor with tag {cbor.tag}")
        fields = frozenlist.FrozenList(list(map(data_from_cbortag, fields)))
        fields.freeze()
        return PlutusConstr(constructor, fields)
    if isinstance(cbor, int):
        return PlutusInteger(cbor)
    if isinstance(cbor, bytes):
        return PlutusByteString(cbor)
    if isinstance(cbor, list):
        entries = frozenlist.FrozenList(list(map(data_from_cbortag, cbor)))
        entries.freeze()
        return PlutusList(entries)
    if isinstance(cbor, dict):
        return PlutusMap(
            frozendict.frozendict(
                {data_from_cbortag(k): data_from_cbortag(v) for k, v in cbor.items()}
            )
        )


def data_from_cbor(cbor: bytes) -> PlutusData:
    raw_datum = cbor2.loads(cbor)
    return data_from_cbortag(raw_datum)


class ConstantType(Enum):
    integer = auto()
    bytestring = auto()
    string = auto()
    unit = auto()
    bool = auto()
    pair = auto()
    list = auto()
    data = auto()


# As found in https://plutonomicon.github.io/plutonomicon/builtin-functions
class BuiltInFun(Enum):
    AddInteger = auto()
    SubtractInteger = auto()
    MultiplyInteger = auto()
    DivideInteger = auto()
    QuotientInteger = auto()
    RemainderInteger = auto()
    ModInteger = auto()
    EqualsInteger = auto()
    LessThanInteger = auto()
    LessThanEqualsInteger = auto()
    AppendByteString = auto()
    ConsByteString = auto()
    SliceByteString = auto()
    LengthOfByteString = auto()
    IndexByteString = auto()
    EqualsByteString = auto()
    LessThanByteString = auto()
    LessThanEqualsByteString = auto()
    Sha2_256 = auto()
    Sha3_256 = auto()
    Blake2b_256 = auto()
    VerifySignature = auto()
    VerifyEd25519Signature = auto()
    VerifyEcdsaSecp256k1Signature = auto()
    VerifySchnorrSecp256k1Signature = auto()
    AppendString = auto()
    EqualsString = auto()
    EncodeUtf8 = auto()
    DecodeUtf8 = auto()
    IfThenElse = auto()
    ChooseUnit = auto()
    Trace = auto()
    FstPair = auto()
    SndPair = auto()
    ChooseList = auto()
    MkCons = auto()
    HeadList = auto()
    TailList = auto()
    NullList = auto()
    ChooseData = auto()
    ConstrData = auto()
    MapData = auto()
    ListData = auto()
    IData = auto()
    BData = auto()
    UnConstrData = auto()
    UnMapData = auto()
    UnListData = auto()
    UnIData = auto()
    UnBData = auto()
    EqualsData = auto()
    MkPairData = auto()
    MkNilData = auto()
    MkNilPairData = auto()


def _IfThenElse(i, t, e):
    assert isinstance(
        i, BuiltinBool
    ), "Trying to compute ifthenelse with non-builtin-bool"
    return t if i.value else e


def _ChooseData(d, v, w, x, y, z):
    if isinstance(d, PlutusConstr):
        return v
    if isinstance(d, PlutusMap):
        return w
    if isinstance(d, PlutusList):
        return x
    if isinstance(d, PlutusInteger):
        return y
    if isinstance(d, PlutusByteString):
        return z


def verify_ed25519(pk: BuiltinByteString, m: BuiltinByteString, s: BuiltinByteString):
    assert len(pk.value) == 32, "Ed25519S PublicKey should be 32 bytes"
    assert len(s.value) == 64, "Ed25519S Signature should be 64 bytes"
    try:
        BIP32ED25519PublicKey(pk.value[:32], pk.value[32:]).verify(s.value, m.value)
        return BuiltinBool(True)
    except nacl.exceptions.BadSignatureError:
        return BuiltinBool(False)


def verify_ecdsa_secp256k1(
    pk: BuiltinByteString, m: BuiltinByteString, s: BuiltinByteString
):
    raise NotImplementedError("EcdsaSecp256k1 not implemented yet")


def verify_schnorr_secp256k1(
    pk: BuiltinByteString, m: BuiltinByteString, s: BuiltinByteString
):
    raise NotImplementedError("SchnorrSecp256k1 not implemented yet")


def _quot(a, b):
    return a // b if (a * b > BuiltinInteger(0)).value else (a + (-a % b)) // b


def _TailList(xs: BuiltinList):
    if xs.values == []:
        raise RuntimeError("Can not tailList on an empty list")
    return xs[1:]


BuiltInFunEvalMap = {
    BuiltInFun.AddInteger: lambda x, y: x + y,
    BuiltInFun.SubtractInteger: lambda x, y: x - y,
    BuiltInFun.MultiplyInteger: lambda x, y: x * y,
    # round towards -inf
    BuiltInFun.DivideInteger: lambda x, y: x // y,
    # round towards 0
    BuiltInFun.QuotientInteger: _quot,
    # (x `quot` y)*y + (x `rem` y) == x
    BuiltInFun.RemainderInteger: lambda x, y: x - _quot(x, y) * y,
    # (x `div` y)*y + (x `mod` y) == x
    BuiltInFun.ModInteger: lambda x, y: x % y,
    BuiltInFun.EqualsInteger: lambda x, y: x == y,
    BuiltInFun.LessThanInteger: lambda x, y: x < y,
    BuiltInFun.LessThanEqualsInteger: lambda x, y: x <= y,
    BuiltInFun.AppendByteString: lambda x, y: x + y,
    BuiltInFun.ConsByteString: lambda x, y: BuiltinByteString(bytes([x.value])) + y,
    BuiltInFun.SliceByteString: lambda x, y, z: BuiltinByteString(
        z.value[max(x.value, 0) :][: max(y.value, 0)]
    ),
    BuiltInFun.LengthOfByteString: lambda x: BuiltinInteger(len(x.value)),
    BuiltInFun.IndexByteString: lambda x, y: x[y],
    BuiltInFun.EqualsByteString: lambda x, y: x == y,
    BuiltInFun.LessThanByteString: lambda x, y: x < y,
    BuiltInFun.LessThanEqualsByteString: lambda x, y: x <= y,
    BuiltInFun.Sha2_256: lambda x: BuiltinByteString(hashlib.sha256(x.value).digest()),
    BuiltInFun.Sha3_256: lambda x: BuiltinByteString(
        hashlib.sha3_256(x.value).digest()
    ),
    BuiltInFun.Blake2b_256: lambda x: BuiltinByteString(
        hashlib.blake2b(x.value, digest_size=32).digest()
    ),
    BuiltInFun.VerifySignature: verify_ed25519,
    BuiltInFun.VerifyEd25519Signature: verify_ed25519,
    # TODO how to emulate this?
    BuiltInFun.VerifyEcdsaSecp256k1Signature: lambda pk, m, s: BuiltinBool(True),
    BuiltInFun.VerifySchnorrSecp256k1Signature: lambda pk, m, s: BuiltinBool(True),
    BuiltInFun.AppendString: lambda x, y: x + y,
    BuiltInFun.EqualsString: lambda x, y: x == y,
    BuiltInFun.EncodeUtf8: lambda x: BuiltinByteString(x.value.encode("utf8")),
    BuiltInFun.DecodeUtf8: lambda x: BuiltinString(x.value.decode("utf8")),
    BuiltInFun.IfThenElse: _IfThenElse,
    BuiltInFun.ChooseUnit: lambda x, y: y,
    BuiltInFun.Trace: lambda x, y: print(x.value) or y,
    BuiltInFun.FstPair: lambda x: x[0],
    BuiltInFun.SndPair: lambda x: x[1],
    BuiltInFun.ChooseList: lambda l, x, y: x
    if l == BuiltinList([], l.sample_value)
    else y,
    BuiltInFun.MkCons: lambda e, l: BuiltinList([e]) + l,
    BuiltInFun.HeadList: lambda l: l[0],
    BuiltInFun.TailList: _TailList,
    BuiltInFun.NullList: lambda l: BuiltinBool(l == BuiltinList([], l.sample_value)),
    BuiltInFun.ChooseData: _ChooseData,
    BuiltInFun.ConstrData: lambda x, y: PlutusConstr(x.value, y.values),
    BuiltInFun.MapData: lambda x: PlutusMap({p.l_value: p.r_value for p in x.values}),
    BuiltInFun.ListData: lambda x: PlutusList(x.values),
    BuiltInFun.IData: lambda x: PlutusInteger(x.value),
    BuiltInFun.BData: lambda x: PlutusByteString(x.value),
    BuiltInFun.UnConstrData: lambda x: BuiltinPair(
        BuiltinInteger(x.constructor), BuiltinList(x.fields, PlutusData())
    ),
    BuiltInFun.UnMapData: lambda x: BuiltinList(
        [BuiltinPair(k, v) for k, v in x.value.items()],
        BuiltinPair(PlutusData(), PlutusData()),
    ),
    BuiltInFun.UnListData: lambda x: BuiltinList(x.value, PlutusData()),
    BuiltInFun.UnIData: lambda x: BuiltinInteger(x.value),
    BuiltInFun.UnBData: lambda x: BuiltinByteString(x.value),
    BuiltInFun.EqualsData: lambda x, y: BuiltinBool(x == y),
    BuiltInFun.MkPairData: lambda x, y: BuiltinPair(x, y),
    BuiltInFun.MkNilData: lambda _: BuiltinList([], PlutusData()),
    BuiltInFun.MkNilPairData: lambda _: BuiltinList(
        [], BuiltinPair(PlutusData(), PlutusData())
    ),
}

BuiltInFunForceMap = defaultdict(int)
BuiltInFunForceMap.update(
    {
        BuiltInFun.IfThenElse: 1,
        BuiltInFun.ChooseUnit: 1,
        BuiltInFun.Trace: 1,
        BuiltInFun.FstPair: 2,
        BuiltInFun.SndPair: 2,
        BuiltInFun.ChooseList: 2,
        BuiltInFun.MkCons: 1,
        BuiltInFun.HeadList: 1,
        BuiltInFun.TailList: 1,
        BuiltInFun.NullList: 1,
        BuiltInFun.ChooseData: 1,
    }
)


@dataclass
class Program(AST):
    version: str
    term: AST
    _fields = ["term"]

    def eval(self, context, state):
        return self.term.eval(context, state)

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        return f"(program {self.version} {self.term.dumps(dialect=dialect)})"


@dataclass
class Variable(AST):
    name: str

    def eval(self, context, state):
        try:
            return Return(context, state[self.name])
        except KeyError as e:
            _LOGGER.error(
                f"Access to uninitialized variable {self.name} in {self.dumps()}"
            )
            raise e

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        return self.name


@dataclass
class BoundStateLambda(AST):
    var_name: str
    term: AST
    state: frozendict.frozendict
    _fields = ["term"]

    def eval(self, context, state):
        return Return(
            context,
            BoundStateLambda(self.var_name, self.term, self.state | state),
        )

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        s = f"(lam {self.var_name} {self.term.dumps(dialect=dialect)})"
        for k, v in reversed(self.state.items()):
            s = f"[(lam {k} {s}) {v.dumps(dialect=dialect)}]"
        return s


@dataclass
class Lambda(BoundStateLambda):
    var_name: str
    term: AST
    state: frozendict.frozendict = dataclasses.field(
        default_factory=frozendict.frozendict
    )


@dataclass
class BoundStateDelay(AST):
    term: AST
    state: frozendict.frozendict
    _fields = ["term"]

    def eval(self, context, state):
        return Return(context, BoundStateDelay(self.term, self.state | state))

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        return f"(delay {self.term.dumps(dialect=dialect)})"


@dataclass
class Delay(BoundStateDelay):
    term: AST
    state: frozendict.frozendict = dataclasses.field(
        default_factory=frozendict.frozendict
    )


@dataclass
class Force(AST):
    term: AST
    _fields = ["term"]

    def eval(self, context, state):
        return Compute(
            FrameForce(
                context,
            ),
            state,
            self.term,
        )

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        return f"(force {self.term.dumps(dialect=dialect)})"


@dataclass
class ForcedBuiltIn(AST):
    builtin: BuiltInFun
    applied_forces: int
    bound_arguments: List[AST]

    def eval(self, context, state):
        return Return(context, self)

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        if len(self.bound_arguments):
            return Apply(
                ForcedBuiltIn(
                    self.builtin, self.applied_forces, self.bound_arguments[:-1]
                ),
                self.bound_arguments[-1],
            ).dumps(dialect=dialect)
        if self.applied_forces > 0:
            return Force(
                ForcedBuiltIn(
                    self.builtin, self.applied_forces - 1, self.bound_arguments
                )
            ).dumps(dialect=dialect)
        return f"(builtin {self.builtin.name[0].lower()}{self.builtin.name[1:]})"


@dataclass
class BuiltIn(ForcedBuiltIn):
    builtin: BuiltInFun
    applied_forces: int = dataclasses.field(default=0)
    bound_arguments: list = dataclasses.field(default_factory=lambda: [])


@dataclass
class Error(AST):
    def eval(self, context, state):
        raise RuntimeError(f"Execution called Error")

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        return f"(error)"


@dataclass
class Apply(AST):
    f: AST
    x: AST
    _fields = ["f", "x"]

    def eval(self, context, state):
        return Compute(
            FrameApplyArg(
                state,
                self.x,
                context,
            ),
            state,
            self.f,
        )

    def dumps(self, dialect=UPLCDialect.Aiken) -> str:
        return f"[{self.f.dumps(dialect=dialect)} {self.x.dumps(dialect=dialect)}]"



import datetime
import unittest
import hypothesis
from hypothesis import strategies as hst
import frozenlist as fl

from .. import *
from ..optimizer import pre_evaluation
from ..transformer import unique_variables
from ..ast import *
from .. import lexer


def frozenlist(l):
    l = fl.FrozenList(l)
    l.freeze()
    return l


pos_int = hst.integers(min_value=0)


uplc_data_integer = hst.builds(PlutusInteger, hst.integers())
uplc_data_bytestring = hst.builds(PlutusByteString, hst.binary())


def rec_data_strategies(uplc_data):
    uplc_data_list = hst.builds(
        lambda x: PlutusList(frozenlist(x)), hst.lists(uplc_data)
    )
    uplc_data_constr = hst.builds(
        lambda x, y: PlutusConstr(x, frozenlist(y)),
        hst.integers(),
        hst.lists(uplc_data),
    )
    uplc_data_map = hst.builds(
        PlutusMap,
        hst.dictionaries(
            hst.one_of(
                uplc_data_integer, uplc_data_bytestring
            ),  # TODO technically constr is legal too, but causes hashing error
            uplc_data,
            dict_class=frozendict.frozendict,
        ),
    )
    return hst.one_of(uplc_data_map, uplc_data_list, uplc_data_constr)


uplc_data = hst.recursive(
    hst.one_of(uplc_data_bytestring, uplc_data_integer),
    rec_data_strategies,
    max_leaves=4,
)
uplc_builtin_boolean = hst.builds(BuiltinBool, hst.booleans())
uplc_builtin_integer = hst.builds(BuiltinInteger, hst.integers())
uplc_builtin_bytestring = hst.builds(BuiltinByteString, hst.binary())
uplc_builtin_string = hst.builds(
    BuiltinString, hst.from_regex(r'([^\n\r"]|\\")*', fullmatch=True)
)
uplc_builtin_unit = hst.just(BuiltinUnit())


def rec_const_strategies(uplc_constant):
    uplc_builtin_pair = hst.builds(BuiltinPair, uplc_constant, uplc_constant)
    uplc_builtin_list = hst.builds(
        lambda x, y: BuiltinList(frozenlist([x] * y), x),
        uplc_constant,
        hst.integers(min_value=0, max_value=10),
    )
    return hst.one_of(uplc_builtin_list, uplc_builtin_pair)


uplc_constant = hst.recursive(
    hst.one_of(
        uplc_builtin_unit,
        uplc_builtin_string,
        uplc_builtin_bytestring,
        uplc_builtin_integer,
        uplc_builtin_boolean,
        uplc_data,
    ),
    rec_const_strategies,
    max_leaves=4,
)
uplc_error = hst.just(Error())
uplc_name = hst.from_regex(r"[a-z_~'][\w~!'#]*", fullmatch=True)
uplc_builtin_fun = hst.builds(BuiltIn, hst.sampled_from(BuiltInFun))
uplc_variable = hst.builds(Variable, uplc_name)


def rec_expr_strategies(uplc_expr):
    uplc_delay = hst.builds(Delay, uplc_expr)
    uplc_force = hst.builds(Force, uplc_expr)
    uplc_apply = hst.builds(Apply, uplc_expr, uplc_expr)
    uplc_lambda = hst.builds(Lambda, uplc_name, uplc_expr)
    return hst.one_of(uplc_lambda, uplc_delay, uplc_force, uplc_apply)


uplc_expr = hst.recursive(
    hst.one_of(uplc_error, uplc_constant, uplc_builtin_fun, uplc_variable),
    rec_expr_strategies,
    max_leaves=10,
)


uplc_version = hst.builds(lambda x, y, z: f"{x}.{y}.{z}", pos_int, pos_int, pos_int)
uplc_program = hst.builds(Program, uplc_version, uplc_expr)


uplc_token = hst.one_of(
    *(hst.from_regex(t, fullmatch=True) for t in lexer.TOKENS.values())
)
uplc_token_concat = hst.recursive(
    uplc_token,
    lambda uplc_token_concat: hst.builds(
        lambda x, y: x.join(y),
        hst.from_regex(r"[\n\r\s]+", fullmatch=True),
        hst.lists(uplc_token_concat, min_size=5),
    ),
)


class HypothesisTests(unittest.TestCase):
    @hypothesis.given(uplc_program, hst.sampled_from(UPLCDialect))
    @hypothesis.settings(max_examples=1000)
    @hypothesis.example(
        Program(version="0.0.0", term=BuiltinByteString(value=b"")), UPLCDialect.Aiken
    )
    @hypothesis.example(
        Program(version="0.0.0", term=BuiltIn(builtin=BuiltInFun.ConstrData)),
        UPLCDialect.Aiken,
    )
    @hypothesis.example(
        Program(version="0.0.0", term=BuiltinInteger(value=0)), UPLCDialect.Aiken
    )
    def test_dumps_parse_roundtrip(self, p, dialect):
        self.assertEqual(parse(dumps(p, dialect)), p)

    @hypothesis.given(uplc_program)
    @hypothesis.settings(max_examples=1000, deadline=datetime.timedelta(seconds=10))
    @hypothesis.example(parse("(program 0.0.0 (lam _ _))"))
    @hypothesis.example(parse("(program 0.0.0 [(lam x0 (lam _ x0)) (con integer 0)])"))
    @hypothesis.example(parse("(program 0.0.0 [(lam _ (delay _)) (con integer 0)])"))
    @hypothesis.example(parse("(program 0.0.0 (lam _ '))"))
    @hypothesis.example(parse("(program 0.0.0 (delay _))"))
    def test_rewrite_no_semantic_change(self, p):
        code = dumps(p)
        try:
            rewrite_p = unique_variables.UniqueVariableTransformer().visit(p)
        except unique_variables.FreeVariableError:
            return
        try:
            res = eval(p)
            res = unique_variables.UniqueVariableTransformer().visit(res)
        except unique_variables.FreeVariableError:
            self.fail(f"Free variable error occurred after evaluation in {code}")
        except Exception as e:
            res = e.__class__
        try:
            rewrite_res = eval(rewrite_p)
            rewrite_res = unique_variables.UniqueVariableTransformer().visit(
                rewrite_res
            )
        except unique_variables.FreeVariableError:
            self.fail(f"Free variable error occurred after evaluation in {code}")
        except Exception as e:
            rewrite_res = e.__class__
        self.assertEqual(
            res,
            rewrite_res,
            f"Two programs evaluate to different results even though only renamed in {code}",
        )

    @hypothesis.given(hst.one_of(hst.text(), uplc_token_concat))
    @hypothesis.settings(max_examples=1000)
    def test_raises_syntaxerror(self, p):
        try:
            parse(p)
        except SyntaxError:
            pass
        except Exception as e:
            self.fail(f"Failed with non-syntaxerror {e}")

    @hypothesis.given(uplc_program)
    @hypothesis.settings(max_examples=1000, deadline=datetime.timedelta(seconds=1))
    @hypothesis.example(
        Program(
            version="0.0.0",
            term=Lambda(
                var_name="_",
                term=Apply(
                    f=BuiltIn(
                        builtin=BuiltInFun.AddInteger,
                        applied_forces=0,
                        bound_arguments=[],
                    ),
                    x=Lambda(
                        var_name="_", term=Error(), state=frozendict.frozendict({})
                    ),
                ),
                state=frozendict.frozendict({}),
            ),
        )
    )
    def test_preeval_no_semantic_change(self, p):
        code = dumps(p)
        orig_p = parse(code).term
        rewrite_p = pre_evaluation.PreEvaluationOptimizer().visit(p).term
        params = []
        try:
            orig_res = orig_p
            for _ in range(100):
                if isinstance(orig_res, BoundStateLambda) or isinstance(
                    orig_res, ForcedBuiltIn
                ):
                    p = BuiltinUnit()
                    params.append(p)
                    orig_res = Apply(orig_res, p)
                if isinstance(orig_res, BoundStateDelay):
                    orig_res = Force(orig_res)
                orig_res = eval(orig_res)
            orig_res = unique_variables.UniqueVariableTransformer().visit(orig_res)
        except unique_variables.FreeVariableError:
            self.fail(f"Free variable error occurred after evaluation in {code}")
        except Exception as e:
            orig_res = e.__class__
        try:
            rewrite_res = rewrite_p
            for _ in range(100):
                if isinstance(rewrite_res, BoundStateLambda) or isinstance(
                    rewrite_res, ForcedBuiltIn
                ):
                    p = params.pop(0)
                    rewrite_res = Apply(rewrite_res, p)
                if isinstance(rewrite_res, BoundStateDelay):
                    rewrite_res = Force(rewrite_res)
                rewrite_res = eval(rewrite_res)
            rewrite_res = unique_variables.UniqueVariableTransformer().visit(
                rewrite_res
            )
        except unique_variables.FreeVariableError:
            self.fail(f"Free variable error occurred after evaluation in {code}")
        except Exception as e:
            rewrite_res = e.__class__
        self.assertEqual(
            orig_res,
            rewrite_res,
            f"Two programs evaluate to different results after optimization in {code}",
        )



import unittest

import rply.parser
from parameterized import parameterized

from .. import *
from ..transformer import unique_variables
from ..optimizer import pre_evaluation
from ..lexer import strip_comments
from ..ast import *

SAMPLE_CONTRACT = p = Program(
    version="0.0.1",
    term=Apply(
        Apply(
            Apply(
                Lambda(
                    var_name="p0",
                    term=Lambda(
                        var_name="p1",
                        term=Lambda(
                            var_name="p2",
                            term=Force(
                                term=Apply(
                                    f=Apply(
                                        f=Apply(
                                            f=Force(
                                                term=BuiltIn(
                                                    builtin=BuiltInFun.IfThenElse,
                                                    applied_forces=0,
                                                    bound_arguments=[],
                                                )
                                            ),
                                            x=Apply(
                                                f=Lambda(
                                                    var_name="s",
                                                    term=Apply(
                                                        f=Lambda(
                                                            var_name="g",
                                                            term=Apply(
                                                                f=Apply(
                                                                    f=Apply(
                                                                        f=Apply(
                                                                            f=Apply(
                                                                                f=Variable(
                                                                                    name="g"
                                                                                ),
                                                                                x=Variable(
                                                                                    name="g"
                                                                                ),
                                                                            ),
                                                                            x=Apply(
                                                                                f=BuiltIn(
                                                                                    builtin=BuiltInFun.UnIData,
                                                                                    applied_forces=0,
                                                                                    bound_arguments=[],
                                                                                ),
                                                                                x=Variable(
                                                                                    name="p0"
                                                                                ),
                                                                            ),
                                                                        ),
                                                                        x=Apply(
                                                                            f=BuiltIn(
                                                                                builtin=BuiltInFun.UnIData,
                                                                                applied_forces=0,
                                                                                bound_arguments=[],
                                                                            ),
                                                                            x=Variable(
                                                                                name="p1"
                                                                            ),
                                                                        ),
                                                                    ),
                                                                    x=Variable(
                                                                        name="p2"
                                                                    ),
                                                                ),
                                                                x=Variable(name="s"),
                                                            ),
                                                            state=frozendict.frozendict(
                                                                {}
                                                            ),
                                                        ),
                                                        x=Force(
                                                            term=Apply(
                                                                f=Apply(
                                                                    f=Apply(
                                                                        f=Lambda(
                                                                            var_name="s",
                                                                            term=Apply(
                                                                                f=Lambda(
                                                                                    var_name="s",
                                                                                    term=Lambda(
                                                                                        var_name="x",
                                                                                        term=Lambda(
                                                                                            var_name="def",
                                                                                            term=Force(
                                                                                                term=Apply(
                                                                                                    f=Apply(
                                                                                                        f=Apply(
                                                                                                            f=Force(
                                                                                                                term=BuiltIn(
                                                                                                                    builtin=BuiltInFun.IfThenElse,
                                                                                                                    applied_forces=0,
                                                                                                                    bound_arguments=[],
                                                                                                                )
                                                                                                            ),
                                                                                                            x=Apply(
                                                                                                                f=Apply(
                                                                                                                    f=BuiltIn(
                                                                                                                        builtin=BuiltInFun.EqualsByteString,
                                                                                                                        applied_forces=0,
                                                                                                                        bound_arguments=[],
                                                                                                                    ),
                                                                                                                    x=Variable(
                                                                                                                        name="x"
                                                                                                                    ),
                                                                                                                ),
                                                                                                                x=BuiltinByteString(
                                                                                                                    value=b"validator"
                                                                                                                ),
                                                                                                            ),
                                                                                                        ),
                                                                                                        x=Delay(
                                                                                                            term=Delay(
                                                                                                                term=Lambda(
                                                                                                                    var_name="f",
                                                                                                                    term=Lambda(
                                                                                                                        var_name="p0",
                                                                                                                        term=Lambda(
                                                                                                                            var_name="p1",
                                                                                                                            term=Lambda(
                                                                                                                                var_name="p2",
                                                                                                                                term=Lambda(
                                                                                                                                    var_name="s",
                                                                                                                                    term=Apply(
                                                                                                                                        f=Lambda(
                                                                                                                                            var_name="s",
                                                                                                                                            term=Apply(
                                                                                                                                                f=Apply(
                                                                                                                                                    f=BuiltIn(
                                                                                                                                                        builtin=BuiltInFun.EqualsInteger,
                                                                                                                                                        applied_forces=0,
                                                                                                                                                        bound_arguments=[],
                                                                                                                                                    ),
                                                                                                                                                    x=Apply(
                                                                                                                                                        f=Lambda(
                                                                                                                                                            var_name="s",
                                                                                                                                                            term=Apply(
                                                                                                                                                                f=Apply(
                                                                                                                                                                    f=BuiltIn(
                                                                                                                                                                        builtin=BuiltInFun.AddInteger,
                                                                                                                                                                        applied_forces=0,
                                                                                                                                                                        bound_arguments=[],
                                                                                                                                                                    ),
                                                                                                                                                                    x=Apply(
                                                                                                                                                                        f=Lambda(
                                                                                                                                                                            var_name="s",
                                                                                                                                                                            term=Force(
                                                                                                                                                                                term=Apply(
                                                                                                                                                                                    f=Apply(
                                                                                                                                                                                        f=Variable(
                                                                                                                                                                                            name="s"
                                                                                                                                                                                        ),
                                                                                                                                                                                        x=BuiltinByteString(
                                                                                                                                                                                            value=b"datum"
                                                                                                                                                                                        ),
                                                                                                                                                                                    ),
                                                                                                                                                                                    x=Delay(
                                                                                                                                                                                        term=Apply(
                                                                                                                                                                                            f=Lambda(
                                                                                                                                                                                                var_name="_",
                                                                                                                                                                                                term=Error(),
                                                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                                                    {}
                                                                                                                                                                                                ),
                                                                                                                                                                                            ),
                                                                                                                                                                                            x=Apply(
                                                                                                                                                                                                f=Apply(
                                                                                                                                                                                                    f=Force(
                                                                                                                                                                                                        term=BuiltIn(
                                                                                                                                                                                                            builtin=BuiltInFun.Trace,
                                                                                                                                                                                                            applied_forces=0,
                                                                                                                                                                                                            bound_arguments=[],
                                                                                                                                                                                                        )
                                                                                                                                                                                                    ),
                                                                                                                                                                                                    x=BuiltinString(
                                                                                                                                                                                                        value="NameError"
                                                                                                                                                                                                    ),
                                                                                                                                                                                                ),
                                                                                                                                                                                                x=BuiltinUnit(),
                                                                                                                                                                                            ),
                                                                                                                                                                                        ),
                                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                                            {}
                                                                                                                                                                                        ),
                                                                                                                                                                                    ),
                                                                                                                                                                                )
                                                                                                                                                                            ),
                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                {}
                                                                                                                                                                            ),
                                                                                                                                                                        ),
                                                                                                                                                                        x=Variable(
                                                                                                                                                                            name="s"
                                                                                                                                                                        ),
                                                                                                                                                                    ),
                                                                                                                                                                ),
                                                                                                                                                                x=Apply(
                                                                                                                                                                    f=Lambda(
                                                                                                                                                                        var_name="s",
                                                                                                                                                                        term=Force(
                                                                                                                                                                            term=Apply(
                                                                                                                                                                                f=Apply(
                                                                                                                                                                                    f=Variable(
                                                                                                                                                                                        name="s"
                                                                                                                                                                                    ),
                                                                                                                                                                                    x=BuiltinByteString(
                                                                                                                                                                                        value=b"redeemer"
                                                                                                                                                                                    ),
                                                                                                                                                                                ),
                                                                                                                                                                                x=Delay(
                                                                                                                                                                                    term=Apply(
                                                                                                                                                                                        f=Lambda(
                                                                                                                                                                                            var_name="_",
                                                                                                                                                                                            term=Error(),
                                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                                {}
                                                                                                                                                                                            ),
                                                                                                                                                                                        ),
                                                                                                                                                                                        x=Apply(
                                                                                                                                                                                            f=Apply(
                                                                                                                                                                                                f=Force(
                                                                                                                                                                                                    term=BuiltIn(
                                                                                                                                                                                                        builtin=BuiltInFun.Trace,
                                                                                                                                                                                                        applied_forces=0,
                                                                                                                                                                                                        bound_arguments=[],
                                                                                                                                                                                                    )
                                                                                                                                                                                                ),
                                                                                                                                                                                                x=BuiltinString(
                                                                                                                                                                                                    value="NameError"
                                                                                                                                                                                                ),
                                                                                                                                                                                            ),
                                                                                                                                                                                            x=BuiltinUnit(),
                                                                                                                                                                                        ),
                                                                                                                                                                                    ),
                                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                                        {}
                                                                                                                                                                                    ),
                                                                                                                                                                                ),
                                                                                                                                                                            )
                                                                                                                                                                        ),
                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                            {}
                                                                                                                                                                        ),
                                                                                                                                                                    ),
                                                                                                                                                                    x=Variable(
                                                                                                                                                                        name="s"
                                                                                                                                                                    ),
                                                                                                                                                                ),
                                                                                                                                                            ),
                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                {}
                                                                                                                                                            ),
                                                                                                                                                        ),
                                                                                                                                                        x=Variable(
                                                                                                                                                            name="s"
                                                                                                                                                        ),
                                                                                                                                                    ),
                                                                                                                                                ),
                                                                                                                                                x=Apply(
                                                                                                                                                    f=Lambda(
                                                                                                                                                        var_name="s",
                                                                                                                                                        term=BuiltinInteger(
                                                                                                                                                            value=42
                                                                                                                                                        ),
                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                            {}
                                                                                                                                                        ),
                                                                                                                                                    ),
                                                                                                                                                    x=Variable(
                                                                                                                                                        name="s"
                                                                                                                                                    ),
                                                                                                                                                ),
                                                                                                                                            ),
                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                {}
                                                                                                                                            ),
                                                                                                                                        ),
                                                                                                                                        x=Apply(
                                                                                                                                            f=Lambda(
                                                                                                                                                var_name="s",
                                                                                                                                                term=Variable(
                                                                                                                                                    name="s"
                                                                                                                                                ),
                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                    {}
                                                                                                                                                ),
                                                                                                                                            ),
                                                                                                                                            x=Lambda(
                                                                                                                                                var_name="x",
                                                                                                                                                term=Lambda(
                                                                                                                                                    var_name="def",
                                                                                                                                                    term=Force(
                                                                                                                                                        term=Apply(
                                                                                                                                                            f=Apply(
                                                                                                                                                                f=Apply(
                                                                                                                                                                    f=Force(
                                                                                                                                                                        term=BuiltIn(
                                                                                                                                                                            builtin=BuiltInFun.IfThenElse,
                                                                                                                                                                            applied_forces=0,
                                                                                                                                                                            bound_arguments=[],
                                                                                                                                                                        )
                                                                                                                                                                    ),
                                                                                                                                                                    x=Apply(
                                                                                                                                                                        f=Apply(
                                                                                                                                                                            f=BuiltIn(
                                                                                                                                                                                builtin=BuiltInFun.EqualsByteString,
                                                                                                                                                                                applied_forces=0,
                                                                                                                                                                                bound_arguments=[],
                                                                                                                                                                            ),
                                                                                                                                                                            x=Variable(
                                                                                                                                                                                name="x"
                                                                                                                                                                            ),
                                                                                                                                                                        ),
                                                                                                                                                                        x=BuiltinByteString(
                                                                                                                                                                            value=b"context"
                                                                                                                                                                        ),
                                                                                                                                                                    ),
                                                                                                                                                                ),
                                                                                                                                                                x=Delay(
                                                                                                                                                                    term=Delay(
                                                                                                                                                                        term=Variable(
                                                                                                                                                                            name="p2"
                                                                                                                                                                        ),
                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                            {}
                                                                                                                                                                        ),
                                                                                                                                                                    ),
                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                        {}
                                                                                                                                                                    ),
                                                                                                                                                                ),
                                                                                                                                                            ),
                                                                                                                                                            x=Delay(
                                                                                                                                                                term=Force(
                                                                                                                                                                    term=Apply(
                                                                                                                                                                        f=Apply(
                                                                                                                                                                            f=Apply(
                                                                                                                                                                                f=Force(
                                                                                                                                                                                    term=BuiltIn(
                                                                                                                                                                                        builtin=BuiltInFun.IfThenElse,
                                                                                                                                                                                        applied_forces=0,
                                                                                                                                                                                        bound_arguments=[],
                                                                                                                                                                                    )
                                                                                                                                                                                ),
                                                                                                                                                                                x=Apply(
                                                                                                                                                                                    f=Apply(
                                                                                                                                                                                        f=BuiltIn(
                                                                                                                                                                                            builtin=BuiltInFun.EqualsByteString,
                                                                                                                                                                                            applied_forces=0,
                                                                                                                                                                                            bound_arguments=[],
                                                                                                                                                                                        ),
                                                                                                                                                                                        x=Variable(
                                                                                                                                                                                            name="x"
                                                                                                                                                                                        ),
                                                                                                                                                                                    ),
                                                                                                                                                                                    x=BuiltinByteString(
                                                                                                                                                                                        value=b"redeemer"
                                                                                                                                                                                    ),
                                                                                                                                                                                ),
                                                                                                                                                                            ),
                                                                                                                                                                            x=Delay(
                                                                                                                                                                                term=Delay(
                                                                                                                                                                                    term=Variable(
                                                                                                                                                                                        name="p1"
                                                                                                                                                                                    ),
                                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                                        {}
                                                                                                                                                                                    ),
                                                                                                                                                                                ),
                                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                                    {}
                                                                                                                                                                                ),
                                                                                                                                                                            ),
                                                                                                                                                                        ),
                                                                                                                                                                        x=Delay(
                                                                                                                                                                            term=Force(
                                                                                                                                                                                term=Apply(
                                                                                                                                                                                    f=Apply(
                                                                                                                                                                                        f=Apply(
                                                                                                                                                                                            f=Force(
                                                                                                                                                                                                term=BuiltIn(
                                                                                                                                                                                                    builtin=BuiltInFun.IfThenElse,
                                                                                                                                                                                                    applied_forces=0,
                                                                                                                                                                                                    bound_arguments=[],
                                                                                                                                                                                                )
                                                                                                                                                                                            ),
                                                                                                                                                                                            x=Apply(
                                                                                                                                                                                                f=Apply(
                                                                                                                                                                                                    f=BuiltIn(
                                                                                                                                                                                                        builtin=BuiltInFun.EqualsByteString,
                                                                                                                                                                                                        applied_forces=0,
                                                                                                                                                                                                        bound_arguments=[],
                                                                                                                                                                                                    ),
                                                                                                                                                                                                    x=Variable(
                                                                                                                                                                                                        name="x"
                                                                                                                                                                                                    ),
                                                                                                                                                                                                ),
                                                                                                                                                                                                x=BuiltinByteString(
                                                                                                                                                                                                    value=b"datum"
                                                                                                                                                                                                ),
                                                                                                                                                                                            ),
                                                                                                                                                                                        ),
                                                                                                                                                                                        x=Delay(
                                                                                                                                                                                            term=Delay(
                                                                                                                                                                                                term=Variable(
                                                                                                                                                                                                    name="p0"
                                                                                                                                                                                                ),
                                                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                                                    {}
                                                                                                                                                                                                ),
                                                                                                                                                                                            ),
                                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                                {}
                                                                                                                                                                                            ),
                                                                                                                                                                                        ),
                                                                                                                                                                                    ),
                                                                                                                                                                                    x=Delay(
                                                                                                                                                                                        term=Force(
                                                                                                                                                                                            term=Apply(
                                                                                                                                                                                                f=Apply(
                                                                                                                                                                                                    f=Apply(
                                                                                                                                                                                                        f=Force(
                                                                                                                                                                                                            term=BuiltIn(
                                                                                                                                                                                                                builtin=BuiltInFun.IfThenElse,
                                                                                                                                                                                                                applied_forces=0,
                                                                                                                                                                                                                bound_arguments=[],
                                                                                                                                                                                                            )
                                                                                                                                                                                                        ),
                                                                                                                                                                                                        x=Apply(
                                                                                                                                                                                                            f=Apply(
                                                                                                                                                                                                                f=BuiltIn(
                                                                                                                                                                                                                    builtin=BuiltInFun.EqualsByteString,
                                                                                                                                                                                                                    applied_forces=0,
                                                                                                                                                                                                                    bound_arguments=[],
                                                                                                                                                                                                                ),
                                                                                                                                                                                                                x=Variable(
                                                                                                                                                                                                                    name="x"
                                                                                                                                                                                                                ),
                                                                                                                                                                                                            ),
                                                                                                                                                                                                            x=BuiltinByteString(
                                                                                                                                                                                                                value=b"validator"
                                                                                                                                                                                                            ),
                                                                                                                                                                                                        ),
                                                                                                                                                                                                    ),
                                                                                                                                                                                                    x=Delay(
                                                                                                                                                                                                        term=Delay(
                                                                                                                                                                                                            term=Variable(
                                                                                                                                                                                                                name="f"
                                                                                                                                                                                                            ),
                                                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                                                {}
                                                                                                                                                                                                            ),
                                                                                                                                                                                                        ),
                                                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                                                            {}
                                                                                                                                                                                                        ),
                                                                                                                                                                                                    ),
                                                                                                                                                                                                ),
                                                                                                                                                                                                x=Delay(
                                                                                                                                                                                                    term=Apply(
                                                                                                                                                                                                        f=Apply(
                                                                                                                                                                                                            f=Variable(
                                                                                                                                                                                                                name="s"
                                                                                                                                                                                                            ),
                                                                                                                                                                                                            x=Variable(
                                                                                                                                                                                                                name="x"
                                                                                                                                                                                                            ),
                                                                                                                                                                                                        ),
                                                                                                                                                                                                        x=Variable(
                                                                                                                                                                                                            name="def"
                                                                                                                                                                                                        ),
                                                                                                                                                                                                    ),
                                                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                                                        {}
                                                                                                                                                                                                    ),
                                                                                                                                                                                                ),
                                                                                                                                                                                            )
                                                                                                                                                                                        ),
                                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                                            {}
                                                                                                                                                                                        ),
                                                                                                                                                                                    ),
                                                                                                                                                                                )
                                                                                                                                                                            ),
                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                {}
                                                                                                                                                                            ),
                                                                                                                                                                        ),
                                                                                                                                                                    )
                                                                                                                                                                ),
                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                    {}
                                                                                                                                                                ),
                                                                                                                                                            ),
                                                                                                                                                        )
                                                                                                                                                    ),
                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                        {}
                                                                                                                                                    ),
                                                                                                                                                ),
                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                    {}
                                                                                                                                                ),
                                                                                                                                            ),
                                                                                                                                        ),
                                                                                                                                    ),
                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                        {}
                                                                                                                                    ),
                                                                                                                                ),
                                                                                                                                state=frozendict.frozendict(
                                                                                                                                    {}
                                                                                                                                ),
                                                                                                                            ),
                                                                                                                            state=frozendict.frozendict(
                                                                                                                                {}
                                                                                                                            ),
                                                                                                                        ),
                                                                                                                        state=frozendict.frozendict(
                                                                                                                            {}
                                                                                                                        ),
                                                                                                                    ),
                                                                                                                    state=frozendict.frozendict(
                                                                                                                        {}
                                                                                                                    ),
                                                                                                                ),
                                                                                                                state=frozendict.frozendict(
                                                                                                                    {}
                                                                                                                ),
                                                                                                            ),
                                                                                                            state=frozendict.frozendict(
                                                                                                                {}
                                                                                                            ),
                                                                                                        ),
                                                                                                    ),
                                                                                                    x=Delay(
                                                                                                        term=Apply(
                                                                                                            f=Apply(
                                                                                                                f=Variable(
                                                                                                                    name="s"
                                                                                                                ),
                                                                                                                x=Variable(
                                                                                                                    name="x"
                                                                                                                ),
                                                                                                            ),
                                                                                                            x=Variable(
                                                                                                                name="def"
                                                                                                            ),
                                                                                                        ),
                                                                                                        state=frozendict.frozendict(
                                                                                                            {}
                                                                                                        ),
                                                                                                    ),
                                                                                                )
                                                                                            ),
                                                                                            state=frozendict.frozendict(
                                                                                                {}
                                                                                            ),
                                                                                        ),
                                                                                        state=frozendict.frozendict(
                                                                                            {}
                                                                                        ),
                                                                                    ),
                                                                                    state=frozendict.frozendict(
                                                                                        {}
                                                                                    ),
                                                                                ),
                                                                                x=Apply(
                                                                                    f=Lambda(
                                                                                        var_name="s",
                                                                                        term=Variable(
                                                                                            name="s"
                                                                                        ),
                                                                                        state=frozendict.frozendict(
                                                                                            {}
                                                                                        ),
                                                                                    ),
                                                                                    x=Apply(
                                                                                        f=Lambda(
                                                                                            var_name="s",
                                                                                            term=Variable(
                                                                                                name="s"
                                                                                            ),
                                                                                            state=frozendict.frozendict(
                                                                                                {}
                                                                                            ),
                                                                                        ),
                                                                                        x=Apply(
                                                                                            f=Lambda(
                                                                                                var_name="s",
                                                                                                term=Variable(
                                                                                                    name="s"
                                                                                                ),
                                                                                                state=frozendict.frozendict(
                                                                                                    {}
                                                                                                ),
                                                                                            ),
                                                                                            x=Apply(
                                                                                                f=Lambda(
                                                                                                    var_name="s",
                                                                                                    term=Variable(
                                                                                                        name="s"
                                                                                                    ),
                                                                                                    state=frozendict.frozendict(
                                                                                                        {}
                                                                                                    ),
                                                                                                ),
                                                                                                x=Apply(
                                                                                                    f=Lambda(
                                                                                                        var_name="s",
                                                                                                        term=Variable(
                                                                                                            name="s"
                                                                                                        ),
                                                                                                        state=frozendict.frozendict(
                                                                                                            {}
                                                                                                        ),
                                                                                                    ),
                                                                                                    x=Apply(
                                                                                                        f=Lambda(
                                                                                                            var_name="s",
                                                                                                            term=Variable(
                                                                                                                name="s"
                                                                                                            ),
                                                                                                            state=frozendict.frozendict(
                                                                                                                {}
                                                                                                            ),
                                                                                                        ),
                                                                                                        x=Apply(
                                                                                                            f=Lambda(
                                                                                                                var_name="s",
                                                                                                                term=Variable(
                                                                                                                    name="s"
                                                                                                                ),
                                                                                                                state=frozendict.frozendict(
                                                                                                                    {}
                                                                                                                ),
                                                                                                            ),
                                                                                                            x=Apply(
                                                                                                                f=Lambda(
                                                                                                                    var_name="s",
                                                                                                                    term=Variable(
                                                                                                                        name="s"
                                                                                                                    ),
                                                                                                                    state=frozendict.frozendict(
                                                                                                                        {}
                                                                                                                    ),
                                                                                                                ),
                                                                                                                x=Apply(
                                                                                                                    f=Lambda(
                                                                                                                        var_name="s",
                                                                                                                        term=Variable(
                                                                                                                            name="s"
                                                                                                                        ),
                                                                                                                        state=frozendict.frozendict(
                                                                                                                            {}
                                                                                                                        ),
                                                                                                                    ),
                                                                                                                    x=Apply(
                                                                                                                        f=Lambda(
                                                                                                                            var_name="s",
                                                                                                                            term=Variable(
                                                                                                                                name="s"
                                                                                                                            ),
                                                                                                                            state=frozendict.frozendict(
                                                                                                                                {}
                                                                                                                            ),
                                                                                                                        ),
                                                                                                                        x=Apply(
                                                                                                                            f=Lambda(
                                                                                                                                var_name="s",
                                                                                                                                term=Variable(
                                                                                                                                    name="s"
                                                                                                                                ),
                                                                                                                                state=frozendict.frozendict(
                                                                                                                                    {}
                                                                                                                                ),
                                                                                                                            ),
                                                                                                                            x=Apply(
                                                                                                                                f=Lambda(
                                                                                                                                    var_name="s",
                                                                                                                                    term=Variable(
                                                                                                                                        name="s"
                                                                                                                                    ),
                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                        {}
                                                                                                                                    ),
                                                                                                                                ),
                                                                                                                                x=Apply(
                                                                                                                                    f=Lambda(
                                                                                                                                        var_name="s",
                                                                                                                                        term=Variable(
                                                                                                                                            name="s"
                                                                                                                                        ),
                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                            {}
                                                                                                                                        ),
                                                                                                                                    ),
                                                                                                                                    x=Apply(
                                                                                                                                        f=Lambda(
                                                                                                                                            var_name="s",
                                                                                                                                            term=Variable(
                                                                                                                                                name="s"
                                                                                                                                            ),
                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                {}
                                                                                                                                            ),
                                                                                                                                        ),
                                                                                                                                        x=Apply(
                                                                                                                                            f=Lambda(
                                                                                                                                                var_name="s",
                                                                                                                                                term=Variable(
                                                                                                                                                    name="s"
                                                                                                                                                ),
                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                    {}
                                                                                                                                                ),
                                                                                                                                            ),
                                                                                                                                            x=Apply(
                                                                                                                                                f=Lambda(
                                                                                                                                                    var_name="s",
                                                                                                                                                    term=Variable(
                                                                                                                                                        name="s"
                                                                                                                                                    ),
                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                        {}
                                                                                                                                                    ),
                                                                                                                                                ),
                                                                                                                                                x=Apply(
                                                                                                                                                    f=Lambda(
                                                                                                                                                        var_name="s",
                                                                                                                                                        term=Variable(
                                                                                                                                                            name="s"
                                                                                                                                                        ),
                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                            {}
                                                                                                                                                        ),
                                                                                                                                                    ),
                                                                                                                                                    x=Apply(
                                                                                                                                                        f=Lambda(
                                                                                                                                                            var_name="s",
                                                                                                                                                            term=Variable(
                                                                                                                                                                name="s"
                                                                                                                                                            ),
                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                {}
                                                                                                                                                            ),
                                                                                                                                                        ),
                                                                                                                                                        x=Apply(
                                                                                                                                                            f=Lambda(
                                                                                                                                                                var_name="s",
                                                                                                                                                                term=Variable(
                                                                                                                                                                    name="s"
                                                                                                                                                                ),
                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                    {}
                                                                                                                                                                ),
                                                                                                                                                            ),
                                                                                                                                                            x=Apply(
                                                                                                                                                                f=Lambda(
                                                                                                                                                                    var_name="s",
                                                                                                                                                                    term=Variable(
                                                                                                                                                                        name="s"
                                                                                                                                                                    ),
                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                        {}
                                                                                                                                                                    ),
                                                                                                                                                                ),
                                                                                                                                                                x=Apply(
                                                                                                                                                                    f=Lambda(
                                                                                                                                                                        var_name="s",
                                                                                                                                                                        term=Variable(
                                                                                                                                                                            name="s"
                                                                                                                                                                        ),
                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                            {}
                                                                                                                                                                        ),
                                                                                                                                                                    ),
                                                                                                                                                                    x=Apply(
                                                                                                                                                                        f=Lambda(
                                                                                                                                                                            var_name="s",
                                                                                                                                                                            term=Variable(
                                                                                                                                                                                name="s"
                                                                                                                                                                            ),
                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                {}
                                                                                                                                                                            ),
                                                                                                                                                                        ),
                                                                                                                                                                        x=Apply(
                                                                                                                                                                            f=Lambda(
                                                                                                                                                                                var_name="s",
                                                                                                                                                                                term=Variable(
                                                                                                                                                                                    name="s"
                                                                                                                                                                                ),
                                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                                    {}
                                                                                                                                                                                ),
                                                                                                                                                                            ),
                                                                                                                                                                            x=Apply(
                                                                                                                                                                                f=Lambda(
                                                                                                                                                                                    var_name="s",
                                                                                                                                                                                    term=Variable(
                                                                                                                                                                                        name="s"
                                                                                                                                                                                    ),
                                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                                        {}
                                                                                                                                                                                    ),
                                                                                                                                                                                ),
                                                                                                                                                                                x=Apply(
                                                                                                                                                                                    f=Lambda(
                                                                                                                                                                                        var_name="s",
                                                                                                                                                                                        term=Variable(
                                                                                                                                                                                            name="s"
                                                                                                                                                                                        ),
                                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                                            {}
                                                                                                                                                                                        ),
                                                                                                                                                                                    ),
                                                                                                                                                                                    x=Apply(
                                                                                                                                                                                        f=Lambda(
                                                                                                                                                                                            var_name="s",
                                                                                                                                                                                            term=Variable(
                                                                                                                                                                                                name="s"
                                                                                                                                                                                            ),
                                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                                {}
                                                                                                                                                                                            ),
                                                                                                                                                                                        ),
                                                                                                                                                                                        x=Apply(
                                                                                                                                                                                            f=Lambda(
                                                                                                                                                                                                var_name="s",
                                                                                                                                                                                                term=Variable(
                                                                                                                                                                                                    name="s"
                                                                                                                                                                                                ),
                                                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                                                    {}
                                                                                                                                                                                                ),
                                                                                                                                                                                            ),
                                                                                                                                                                                            x=Apply(
                                                                                                                                                                                                f=Lambda(
                                                                                                                                                                                                    var_name="s",
                                                                                                                                                                                                    term=Variable(
                                                                                                                                                                                                        name="s"
                                                                                                                                                                                                    ),
                                                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                                                        {}
                                                                                                                                                                                                    ),
                                                                                                                                                                                                ),
                                                                                                                                                                                                x=Apply(
                                                                                                                                                                                                    f=Lambda(
                                                                                                                                                                                                        var_name="s",
                                                                                                                                                                                                        term=Variable(
                                                                                                                                                                                                            name="s"
                                                                                                                                                                                                        ),
                                                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                                                            {}
                                                                                                                                                                                                        ),
                                                                                                                                                                                                    ),
                                                                                                                                                                                                    x=Apply(
                                                                                                                                                                                                        f=Lambda(
                                                                                                                                                                                                            var_name="s",
                                                                                                                                                                                                            term=Variable(
                                                                                                                                                                                                                name="s"
                                                                                                                                                                                                            ),
                                                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                                                {}
                                                                                                                                                                                                            ),
                                                                                                                                                                                                        ),
                                                                                                                                                                                                        x=Apply(
                                                                                                                                                                                                            f=Lambda(
                                                                                                                                                                                                                var_name="s",
                                                                                                                                                                                                                term=Variable(
                                                                                                                                                                                                                    name="s"
                                                                                                                                                                                                                ),
                                                                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                                                                    {}
                                                                                                                                                                                                                ),
                                                                                                                                                                                                            ),
                                                                                                                                                                                                            x=Apply(
                                                                                                                                                                                                                f=Lambda(
                                                                                                                                                                                                                    var_name="s",
                                                                                                                                                                                                                    term=Variable(
                                                                                                                                                                                                                        name="s"
                                                                                                                                                                                                                    ),
                                                                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                                                                        {}
                                                                                                                                                                                                                    ),
                                                                                                                                                                                                                ),
                                                                                                                                                                                                                x=Apply(
                                                                                                                                                                                                                    f=Lambda(
                                                                                                                                                                                                                        var_name="s",
                                                                                                                                                                                                                        term=Variable(
                                                                                                                                                                                                                            name="s"
                                                                                                                                                                                                                        ),
                                                                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                                                                            {}
                                                                                                                                                                                                                        ),
                                                                                                                                                                                                                    ),
                                                                                                                                                                                                                    x=Apply(
                                                                                                                                                                                                                        f=Lambda(
                                                                                                                                                                                                                            var_name="s",
                                                                                                                                                                                                                            term=Variable(
                                                                                                                                                                                                                                name="s"
                                                                                                                                                                                                                            ),
                                                                                                                                                                                                                            state=frozendict.frozendict(
                                                                                                                                                                                                                                {}
                                                                                                                                                                                                                            ),
                                                                                                                                                                                                                        ),
                                                                                                                                                                                                                        x=Apply(
                                                                                                                                                                                                                            f=Lambda(
                                                                                                                                                                                                                                var_name="s",
                                                                                                                                                                                                                                term=Variable(
                                                                                                                                                                                                                                    name="s"
                                                                                                                                                                                                                                ),
                                                                                                                                                                                                                                state=frozendict.frozendict(
                                                                                                                                                                                                                                    {}
                                                                                                                                                                                                                                ),
                                                                                                                                                                                                                            ),
                                                                                                                                                                                                                            x=Apply(
                                                                                                                                                                                                                                f=Lambda(
                                                                                                                                                                                                                                    var_name="s",
                                                                                                                                                                                                                                    term=Variable(
                                                                                                                                                                                                                                        name="s"
                                                                                                                                                                                                                                    ),
                                                                                                                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                                                                                                                        {}
                                                                                                                                                                                                                                    ),
                                                                                                                                                                                                                                ),
                                                                                                                                                                                                                                x=Apply(
                                                                                                                                                                                                                                    f=Lambda(
                                                                                                                                                                                                                                        var_name="s",
                                                                                                                                                                                                                                        term=Variable(
                                                                                                                                                                                                                                            name="s"
                                                                                                                                                                                                                                        ),
                                                                                                                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                                                                                                                            {}
                                                                                                                                                                                                                                        ),
                                                                                                                                                                                                                                    ),
                                                                                                                                                                                                                                    x=Variable(
                                                                                                                                                                                                                                        name="s"
                                                                                                                                                                                                                                    ),
                                                                                                                                                                                                                                ),
                                                                                                                                                                                                                            ),
                                                                                                                                                                                                                        ),
                                                                                                                                                                                                                    ),
                                                                                                                                                                                                                ),
                                                                                                                                                                                                            ),
                                                                                                                                                                                                        ),
                                                                                                                                                                                                    ),
                                                                                                                                                                                                ),
                                                                                                                                                                                            ),
                                                                                                                                                                                        ),
                                                                                                                                                                                    ),
                                                                                                                                                                                ),
                                                                                                                                                                            ),
                                                                                                                                                                        ),
                                                                                                                                                                    ),
                                                                                                                                                                ),
                                                                                                                                                            ),
                                                                                                                                                        ),
                                                                                                                                                    ),
                                                                                                                                                ),
                                                                                                                                            ),
                                                                                                                                        ),
                                                                                                                                    ),
                                                                                                                                ),
                                                                                                                            ),
                                                                                                                        ),
                                                                                                                    ),
                                                                                                                ),
                                                                                                            ),
                                                                                                        ),
                                                                                                    ),
                                                                                                ),
                                                                                            ),
                                                                                        ),
                                                                                    ),
                                                                                ),
                                                                            ),
                                                                            state=frozendict.frozendict(
                                                                                {}
                                                                            ),
                                                                        ),
                                                                        x=Variable(
                                                                            name="s"
                                                                        ),
                                                                    ),
                                                                    x=BuiltinByteString(
                                                                        value=b"validator"
                                                                    ),
                                                                ),
                                                                x=Delay(
                                                                    term=Apply(
                                                                        f=Lambda(
                                                                            var_name="_",
                                                                            term=Error(),
                                                                            state=frozendict.frozendict(
                                                                                {}
                                                                            ),
                                                                        ),
                                                                        x=Apply(
                                                                            f=Apply(
                                                                                f=Force(
                                                                                    term=BuiltIn(
                                                                                        builtin=BuiltInFun.Trace,
                                                                                        applied_forces=0,
                                                                                        bound_arguments=[],
                                                                                    )
                                                                                ),
                                                                                x=BuiltinString(
                                                                                    value="NameError"
                                                                                ),
                                                                            ),
                                                                            x=BuiltinUnit(),
                                                                        ),
                                                                    ),
                                                                    state=frozendict.frozendict(
                                                                        {}
                                                                    ),
                                                                ),
                                                            )
                                                        ),
                                                    ),
                                                    state=frozendict.frozendict({}),
                                                ),
                                                x=Lambda(
                                                    var_name="x",
                                                    term=Lambda(
                                                        var_name="def",
                                                        term=Force(
                                                            term=Apply(
                                                                f=Apply(
                                                                    f=Apply(
                                                                        f=Force(
                                                                            term=BuiltIn(
                                                                                builtin=BuiltInFun.IfThenElse,
                                                                                applied_forces=0,
                                                                                bound_arguments=[],
                                                                            )
                                                                        ),
                                                                        x=Apply(
                                                                            f=Apply(
                                                                                f=BuiltIn(
                                                                                    builtin=BuiltInFun.EqualsByteString,
                                                                                    applied_forces=0,
                                                                                    bound_arguments=[],
                                                                                ),
                                                                                x=Variable(
                                                                                    name="x"
                                                                                ),
                                                                            ),
                                                                            x=BuiltinByteString(
                                                                                value=b"range"
                                                                            ),
                                                                        ),
                                                                    ),
                                                                    x=Delay(
                                                                        term=Delay(
                                                                            term=Lambda(
                                                                                var_name="f",
                                                                                term=Lambda(
                                                                                    var_name="limit",
                                                                                    term=Lambda(
                                                                                        var_name="s",
                                                                                        term=Apply(
                                                                                            f=Apply(
                                                                                                f=Apply(
                                                                                                    f=Lambda(
                                                                                                        var_name="limit",
                                                                                                        term=Lambda(
                                                                                                            var_name="step",
                                                                                                            term=Apply(
                                                                                                                f=Lambda(
                                                                                                                    var_name="g",
                                                                                                                    term=Apply(
                                                                                                                        f=Variable(
                                                                                                                            name="g"
                                                                                                                        ),
                                                                                                                        x=Variable(
                                                                                                                            name="g"
                                                                                                                        ),
                                                                                                                    ),
                                                                                                                    state=frozendict.frozendict(
                                                                                                                        {}
                                                                                                                    ),
                                                                                                                ),
                                                                                                                x=Lambda(
                                                                                                                    var_name="f",
                                                                                                                    term=Lambda(
                                                                                                                        var_name="cur",
                                                                                                                        term=Force(
                                                                                                                            term=Apply(
                                                                                                                                f=Apply(
                                                                                                                                    f=Apply(
                                                                                                                                        f=Force(
                                                                                                                                            term=BuiltIn(
                                                                                                                                                builtin=BuiltInFun.IfThenElse,
                                                                                                                                                applied_forces=0,
                                                                                                                                                bound_arguments=[],
                                                                                                                                            )
                                                                                                                                        ),
                                                                                                                                        x=Apply(
                                                                                                                                            f=Apply(
                                                                                                                                                f=BuiltIn(
                                                                                                                                                    builtin=BuiltInFun.LessThanInteger,
                                                                                                                                                    applied_forces=0,
                                                                                                                                                    bound_arguments=[],
                                                                                                                                                ),
                                                                                                                                                x=Variable(
                                                                                                                                                    name="cur"
                                                                                                                                                ),
                                                                                                                                            ),
                                                                                                                                            x=Variable(
                                                                                                                                                name="limit"
                                                                                                                                            ),
                                                                                                                                        ),
                                                                                                                                    ),
                                                                                                                                    x=Delay(
                                                                                                                                        term=Apply(
                                                                                                                                            f=Apply(
                                                                                                                                                f=Force(
                                                                                                                                                    term=BuiltIn(
                                                                                                                                                        builtin=BuiltInFun.MkCons,
                                                                                                                                                        applied_forces=0,
                                                                                                                                                        bound_arguments=[],
                                                                                                                                                    )
                                                                                                                                                ),
                                                                                                                                                x=Variable(
                                                                                                                                                    name="cur"
                                                                                                                                                ),
                                                                                                                                            ),
                                                                                                                                            x=Apply(
                                                                                                                                                f=Apply(
                                                                                                                                                    f=Variable(
                                                                                                                                                        name="f"
                                                                                                                                                    ),
                                                                                                                                                    x=Variable(
                                                                                                                                                        name="f"
                                                                                                                                                    ),
                                                                                                                                                ),
                                                                                                                                                x=Apply(
                                                                                                                                                    f=Apply(
                                                                                                                                                        f=BuiltIn(
                                                                                                                                                            builtin=BuiltInFun.AddInteger,
                                                                                                                                                            applied_forces=0,
                                                                                                                                                            bound_arguments=[],
                                                                                                                                                        ),
                                                                                                                                                        x=Variable(
                                                                                                                                                            name="cur"
                                                                                                                                                        ),
                                                                                                                                                    ),
                                                                                                                                                    x=Variable(
                                                                                                                                                        name="step"
                                                                                                                                                    ),
                                                                                                                                                ),
                                                                                                                                            ),
                                                                                                                                        ),
                                                                                                                                        state=frozendict.frozendict(
                                                                                                                                            {}
                                                                                                                                        ),
                                                                                                                                    ),
                                                                                                                                ),
                                                                                                                                x=Delay(
                                                                                                                                    term=Apply(
                                                                                                                                        f=BuiltIn(
                                                                                                                                            builtin=BuiltInFun.MkNilData,
                                                                                                                                            applied_forces=0,
                                                                                                                                            bound_arguments=[],
                                                                                                                                        ),
                                                                                                                                        x=BuiltinUnit(),
                                                                                                                                    ),
                                                                                                                                    state=frozendict.frozendict(
                                                                                                                                        {}
                                                                                                                                    ),
                                                                                                                                ),
                                                                                                                            )
                                                                                                                        ),
                                                                                                                        state=frozendict.frozendict(
                                                                                                                            {}
                                                                                                                        ),
                                                                                                                    ),
                                                                                                                    state=frozendict.frozendict(
                                                                                                                        {}
                                                                                                                    ),
                                                                                                                ),
                                                                                                            ),
                                                                                                            state=frozendict.frozendict(
                                                                                                                {}
                                                                                                            ),
                                                                                                        ),
                                                                                                        state=frozendict.frozendict(
                                                                                                            {}
                                                                                                        ),
                                                                                                    ),
                                                                                                    x=Variable(
                                                                                                        name="limit"
                                                                                                    ),
                                                                                                ),
                                                                                                x=BuiltinInteger(
                                                                                                    value=1
                                                                                                ),
                                                                                            ),
                                                                                            x=BuiltinInteger(
                                                                                                value=0
                                                                                            ),
                                                                                        ),
                                                                                        state=frozendict.frozendict(
                                                                                            {}
                                                                                        ),
                                                                                    ),
                                                                                    state=frozendict.frozendict(
                                                                                        {}
                                                                                    ),
                                                                                ),
                                                                                state=frozendict.frozendict(
                                                                                    {}
                                                                                ),
                                                                            ),
                                                                            state=frozendict.frozendict(
                                                                                {}
                                                                            ),
                                                                        ),
                                                                        state=frozendict.frozendict(
                                                                            {}
                                                                        ),
                                                                    ),
                                                                ),
                                                                x=Delay(
                                                                    term=Force(
                                                                        term=Apply(
                                                                            f=Apply(
                                                                                f=Apply(
                                                                                    f=Force(
                                                                                        term=BuiltIn(
                                                                                            builtin=BuiltInFun.IfThenElse,
                                                                                            applied_forces=0,
                                                                                            bound_arguments=[],
                                                                                        )
                                                                                    ),
                                                                                    x=Apply(
                                                                                        f=Apply(
                                                                                            f=BuiltIn(
                                                                                                builtin=BuiltInFun.EqualsByteString,
                                                                                                applied_forces=0,
                                                                                                bound_arguments=[],
                                                                                            ),
                                                                                            x=Variable(
                                                                                                name="x"
                                                                                            ),
                                                                                        ),
                                                                                        x=BuiltinByteString(
                                                                                            value=b"print"
                                                                                        ),
                                                                                    ),
                                                                                ),
                                                                                x=Delay(
                                                                                    term=Delay(
                                                                                        term=Lambda(
                                                                                            var_name="f",
                                                                                            term=Lambda(
                                                                                                var_name="x",
                                                                                                term=Lambda(
                                                                                                    var_name="s",
                                                                                                    term=Apply(
                                                                                                        f=Apply(
                                                                                                            f=Force(
                                                                                                                term=BuiltIn(
                                                                                                                    builtin=BuiltInFun.Trace,
                                                                                                                    applied_forces=0,
                                                                                                                    bound_arguments=[],
                                                                                                                )
                                                                                                            ),
                                                                                                            x=Variable(
                                                                                                                name="x"
                                                                                                            ),
                                                                                                        ),
                                                                                                        x=Apply(
                                                                                                            f=Apply(
                                                                                                                f=BuiltIn(
                                                                                                                    builtin=BuiltInFun.ConstrData,
                                                                                                                    applied_forces=0,
                                                                                                                    bound_arguments=[],
                                                                                                                ),
                                                                                                                x=BuiltinInteger(
                                                                                                                    value=0
                                                                                                                ),
                                                                                                            ),
                                                                                                            x=Apply(
                                                                                                                f=BuiltIn(
                                                                                                                    builtin=BuiltInFun.MkNilData,
                                                                                                                    applied_forces=0,
                                                                                                                    bound_arguments=[],
                                                                                                                ),
                                                                                                                x=BuiltinUnit(),
                                                                                                            ),
                                                                                                        ),
                                                                                                    ),
                                                                                                    state=frozendict.frozendict(
                                                                                                        {}
                                                                                                    ),
                                                                                                ),
                                                                                                state=frozendict.frozendict(
                                                                                                    {}
                                                                                                ),
                                                                                            ),
                                                                                            state=frozendict.frozendict(
                                                                                                {}
                                                                                            ),
                                                                                        ),
                                                                                        state=frozendict.frozendict(
                                                                                            {}
                                                                                        ),
                                                                                    ),
                                                                                    state=frozendict.frozendict(
                                                                                        {}
                                                                                    ),
                                                                                ),
                                                                            ),
                                                                            x=Delay(
                                                                                term=Apply(
                                                                                    f=Apply(
                                                                                        f=Lambda(
                                                                                            var_name="x",
                                                                                            term=Lambda(
                                                                                                var_name="def",
                                                                                                term=Variable(
                                                                                                    name="def"
                                                                                                ),
                                                                                                state=frozendict.frozendict(
                                                                                                    {}
                                                                                                ),
                                                                                            ),
                                                                                            state=frozendict.frozendict(
                                                                                                {}
                                                                                            ),
                                                                                        ),
                                                                                        x=Variable(
                                                                                            name="x"
                                                                                        ),
                                                                                    ),
                                                                                    x=Variable(
                                                                                        name="def"
                                                                                    ),
                                                                                ),
                                                                                state=frozendict.frozendict(
                                                                                    {}
                                                                                ),
                                                                            ),
                                                                        )
                                                                    ),
                                                                    state=frozendict.frozendict(
                                                                        {}
                                                                    ),
                                                                ),
                                                            )
                                                        ),
                                                        state=frozendict.frozendict({}),
                                                    ),
                                                    state=frozendict.frozendict({}),
                                                ),
                                            ),
                                        ),
                                        x=Delay(
                                            term=BuiltinUnit(),
                                            state=frozendict.frozendict({}),
                                        ),
                                    ),
                                    x=Delay(
                                        term=Apply(
                                            f=Lambda(
                                                var_name="_",
                                                term=Error(),
                                                state=frozendict.frozendict({}),
                                            ),
                                            x=Apply(
                                                f=Apply(
                                                    f=Force(
                                                        term=BuiltIn(
                                                            builtin=BuiltInFun.Trace,
                                                            applied_forces=0,
                                                            bound_arguments=[],
                                                        )
                                                    ),
                                                    x=BuiltinString(
                                                        value="ValidationError"
                                                    ),
                                                ),
                                                x=BuiltinUnit(),
                                            ),
                                        ),
                                        state=frozendict.frozendict({}),
                                    ),
                                )
                            ),
                            state=frozendict.frozendict({}),
                        ),
                        state=frozendict.frozendict({}),
                    ),
                    state=frozendict.frozendict({}),
                ),
                PlutusInteger(20),
            ),
            PlutusInteger(22),
        ),
        BuiltinUnit(),
    ),
)


class MiscTest(unittest.TestCase):
    def test_simple_contract(self):
        p = SAMPLE_CONTRACT
        # should not raise
        d = dumps(p)
        # should not raise
        parse(d)
        # should not raise
        r = eval(p)
        self.assertEqual(r, BuiltinUnit())

    def test_unpack_plutus_data(self):
        p = Program(
            "0.0.1",
            Apply(
                BuiltIn(BuiltInFun.UnConstrData),
                data_from_cbor(
                    bytes.fromhex(
                        "d8799fd8799fd8799fd8799f581ce3a0254c00994f731550f81239f12a60c9fd3ce9b9b191543152ec22ffd8799fd8799fd8799f581cb1bec305ddc80189dac8b628ee0adfbe5245c53b84e678ed7ec23d75ffffffff581ce3a0254c00994f731550f81239f12a60c9fd3ce9b9b191543152ec221b0000018bcfe56800d8799fd8799f4040ffd8799f581cdda5fdb1002f7389b33e036b6afee82a8189becb6cba852e8b79b4fb480014df1047454e53ffffffd8799fd87a801a00989680ffff"
                    )
                ),
            ),
        )
        # should not raise anything
        d = dumps(p)
        # should not raise anything
        parse(d)
        r = eval(p)
        # should not raise anything
        r.dumps()
        self.assertEqual(
            r,
            BuiltinPair(
                l_value=BuiltinInteger(value=0),
                r_value=BuiltinList(
                    values=[
                        PlutusConstr(
                            constructor=0,
                            fields=[
                                PlutusConstr(
                                    constructor=0,
                                    fields=[
                                        PlutusConstr(
                                            constructor=0,
                                            fields=[
                                                PlutusByteString(
                                                    value=b'\xe3\xa0%L\x00\x99Os\x15P\xf8\x129\xf1*`\xc9\xfd<\xe9\xb9\xb1\x91T1R\xec"'
                                                )
                                            ],
                                        ),
                                        PlutusConstr(
                                            constructor=0,
                                            fields=[
                                                PlutusConstr(
                                                    constructor=0,
                                                    fields=[
                                                        PlutusConstr(
                                                            constructor=0,
                                                            fields=[
                                                                PlutusByteString(
                                                                    value=b"\xb1\xbe\xc3\x05\xdd\xc8\x01\x89\xda\xc8\xb6(\xee\n\xdf\xbeRE\xc5;\x84\xe6x\xed~\xc2=u"
                                                                )
                                                            ],
                                                        )
                                                    ],
                                                )
                                            ],
                                        ),
                                    ],
                                ),
                                PlutusByteString(
                                    value=b'\xe3\xa0%L\x00\x99Os\x15P\xf8\x129\xf1*`\xc9\xfd<\xe9\xb9\xb1\x91T1R\xec"'
                                ),
                                PlutusInteger(value=1700000000000),
                                PlutusConstr(
                                    constructor=0,
                                    fields=[
                                        PlutusConstr(
                                            constructor=0,
                                            fields=[
                                                PlutusByteString(value=b""),
                                                PlutusByteString(value=b""),
                                            ],
                                        ),
                                        PlutusConstr(
                                            constructor=0,
                                            fields=[
                                                PlutusByteString(
                                                    value=b"\xdd\xa5\xfd\xb1\x00/s\x89\xb3>\x03kj\xfe\xe8*\x81\x89\xbe\xcbl\xba\x85.\x8by\xb4\xfb"
                                                ),
                                                PlutusByteString(
                                                    value=b"\x00\x14\xdf\x10GENS"
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ],
                        ),
                        PlutusConstr(
                            constructor=0,
                            fields=[
                                PlutusConstr(constructor=1, fields=[]),
                                PlutusInteger(value=10000000),
                            ],
                        ),
                    ]
                ),
            ),
        )

    def test_parse(self):
        p = parse(
            """
(program
  1.0.0
  [ [ [ (force (delay [(lam i_0 (con integer 2)) (con bytestring #02)])) (builtin addInteger) ] (error) ] (con pair<list<integer>,unit> [[],()]) ]
)
        """
        )
        print(dumps(p))

    def test_parse_plutus(self):
        p = parse(
            """(program 1.0.0 (con (pair (list integer) bytestring) ([1], #01)))"""
        )
        print(dumps(p, dialect=UPLCDialect.Plutus))

    def test_simple_contract_rewrite(self):
        p = SAMPLE_CONTRACT
        # should not raise
        p = unique_variables.UniqueVariableTransformer().visit(p)
        # should not raise
        d = dumps(p)
        # should not raise
        parse(d)
        r = eval(p)
        self.assertEqual(r, BuiltinUnit())

    @parameterized.expand(
        [
            (
                "-- here is a comment \n and here is normal text",
                "\n and here is normal text",
            ),
            ("-- here is a comment ", ""),
        ]
    )
    def test_strip_comments(self, input, expected):
        self.assertEqual(strip_comments(input), expected)

    @parameterized.expand(
        [
            ("""(program 1.0.0 (con data 123)""",),
            ("""(program 1.0.0 (con data { ByteString #00})""",),
        ]
    )
    def test_parse_fail(self, i):
        try:
            parse(i)
            self.fail("Unexpextedly passed")
        except SyntaxError:
            pass

    def test_simple_contract_optimize(self):
        p = SAMPLE_CONTRACT
        # should not raise
        p = pre_evaluation.PreEvaluationOptimizer().visit(p)
        # should not raise
        d = dumps(p)
        # should not raise
        parse(d)
        r = eval(p)
        self.assertEqual(r, BuiltinUnit())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

import uplc

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="uplc",
    version=uplc.__version__,
    description="Python implementation of untyped plutus language core",
    author=uplc.__author__,
    author_email=uplc.__author_email__,
    url=uplc.__url__,
    py_modules=["uplc"],
    packages=find_packages(),
    install_requires=[
        "frozendict==2.3.4",
        "cbor2==5.4.6",
        "frozenlist==1.3.3",
        "rply==0.7.8",
        "pyaiken==0.3.0",
        "pycardano==0.7.2",
    ],
    tests_require=[
        "hypothesis==6.62.0",
        "parameterized==0.8.1",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=uplc.__license__,
    classifiers=[
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Assemblers",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="python cardano smart contract blockchain verification haskell",
    python_requires=">=3",
    test_suite="uplc.tests",
    entry_points={
        "console_scripts": ["uplc=uplc.__main__:main"],
    },
)



from copy import copy

from ..util import NodeTransformer, eval
from ..ast import Program, AST

"""
Optimizes code by pre-evaluating each subterm
If it throws an error, assume it is not safe to pre-evaluate and don't replace, otherwise replace by result.
"""


class PreEvaluationOptimizer(NodeTransformer):
    def visit_Program(self, node: Program) -> Program:
        return Program(version=node.version, term=self.visit(node.term))

    def generic_visit(self, node: AST) -> AST:
        try:
            nc = eval(node)
        except Exception as e:
            nc = node
        return super().generic_visit(nc)



import ast as python_ast
import re

from rply import ParserGenerator
import rply
from . import lexer, ast


class Parser:
    def __init__(self):
        self.pg = ParserGenerator(lexer.TOKENS.keys())

        @self.pg.production(
            "program : PAREN_OPEN PROGRAM version expression PAREN_CLOSE"
        )
        def program(p):
            return ast.Program(p[2], p[3])

        @self.pg.production("version : NUMBER DOT NUMBER DOT NUMBER")
        def version(p):
            return f"{int(p[0].value)}.{int(p[2].value)}.{int(p[4].value)}"

        @self.pg.production(
            "expression : PAREN_OPEN LAMBDA NAME expression PAREN_CLOSE"
        )
        def expression(p):
            return ast.Lambda(p[2].value, p[3])

        @self.pg.production("expression : NAME")
        def expression(p):
            return ast.Variable(p[0].value)

        @self.pg.production("expression : PAREN_OPEN FORCE expression PAREN_CLOSE")
        def force(p):
            return ast.Force(p[2])

        @self.pg.production("expression : PAREN_OPEN DELAY expression PAREN_CLOSE")
        def delay(p):
            return ast.Delay(p[2])

        @self.pg.production("expression : PAREN_OPEN ERROR PAREN_CLOSE")
        def error(p):
            return ast.Error()

        @self.pg.production("expression : PAREN_OPEN BUILTIN NAME PAREN_CLOSE")
        def builtin(p):
            bfn = p[2].value.lower()
            correct_bfn = None
            for e in ast.BuiltInFun:
                if e.name.lower() == bfn:
                    correct_bfn = e
            if correct_bfn is None:
                raise SyntaxError(f"Unknown builtin function {bfn}")
            return ast.BuiltIn(correct_bfn)

        @self.pg.production(
            "expression : BRACK_OPEN expression expression_list BRACK_CLOSE"
        )
        def delay(p):
            res = p[1]
            for e in p[2]:
                res = ast.Apply(res, e)
            return res

        @self.pg.production("expression_list : expression")
        def delay(p):
            return [p[0]]

        @self.pg.production("expression_list : expression expression_list")
        def delay(p):
            return [p[0]] + p[1]

        @self.pg.production("builtintype : NAME")
        def builtintype(p):
            name = p[0].value
            if name == "integer":
                return ast.BuiltinInteger(0)
            if name == "bytestring":
                return ast.BuiltinByteString(b"")
            if name == "string":
                return ast.BuiltinString("")
            if name == "bool":
                return ast.BuiltinBool(False)
            if name == "unit":
                return ast.BuiltinUnit()
            if name == "data":
                return ast.PlutusData()
            raise SyntaxError(f"Unknown builtin type {name}")

        @self.pg.production("builtintype : NAME CARET_OPEN builtintype CARET_CLOSE")
        def builtintype(p):
            # the Aiken dialect
            name = p[0].value
            if name == "list":
                return ast.BuiltinList([], p[2])
            raise SyntaxError(f"Unknown builtin type {name}")

        @self.pg.production("builtintype : PAREN_OPEN NAME builtintype PAREN_CLOSE")
        def builtintype(p):
            # the Plutus dialect
            name = p[1].value
            if name == "list":
                return ast.BuiltinList([], p[2])
            raise SyntaxError(f"Unknown builtin type {name}")

        @self.pg.production(
            "builtintype : NAME CARET_OPEN builtintype COMMA builtintype CARET_CLOSE"
        )
        def builtintype(p):
            # the Aiken dialect
            name = p[0].value
            if name == "pair":
                return ast.BuiltinPair(p[2], p[4])
            raise SyntaxError(f"Unknown builtin type {name}")

        @self.pg.production(
            "builtintype : PAREN_OPEN NAME builtintype builtintype PAREN_CLOSE"
        )
        def builtintype(p):
            # the Plutus dialect
            name = p[1].value
            if name == "pair":
                return ast.BuiltinPair(p[2], p[3])
            raise SyntaxError(f"Unknown builtin type {name}")

        @self.pg.production(
            "expression : PAREN_OPEN CON builtintype builtinvalue PAREN_CLOSE"
        )
        def constant(p):
            typ = p[2]
            val = p[3]
            return wrap_builtin_type(typ, val)

        @self.pg.production("builtinvalue : HEX")
        def expression(p):
            return bytes.fromhex(p[0].value[1:])

        @self.pg.production("builtinvalue : NUMBER")
        def expression(p):
            return int(p[0].value)

        @self.pg.production("builtinvalue : TEXT")
        def expression(p):
            return python_ast.literal_eval(p[0].value)

        @self.pg.production("builtinvalue : PAREN_OPEN PAREN_CLOSE")
        def expression(p):
            return None

        @self.pg.production("builtinvalue : NAME")
        def expression(p):
            assert p[0].value in ("True", "False"), f"Invalid boolean constant {p}"
            return p[0].value == "True"

        @self.pg.production("builtinvaluelist : builtinvalue COMMA builtinvaluelist ")
        def expression(p):
            return [p[0]] + p[2]

        @self.pg.production("builtinvaluelist : builtinvalue builtinvaluelist ")
        def expression(p):
            return [p[0]] + p[1]

        @self.pg.production("builtinvaluelist : BRACK_CLOSE ")
        def expression(p):
            return []

        @self.pg.production("builtinvalue : BRACK_OPEN builtinvaluelist")
        def expression(p):
            # the Aiken dialect for pair and list values
            # and the Plutus dialect for list values
            return p[1]

        @self.pg.production(
            "builtinvalue : PAREN_OPEN builtinvalue COMMA builtinvalue PAREN_CLOSE"
        )
        def expression(p):
            # and the Plutus dialect for pairs
            return (p[1], p[3])

    def get_parser(self):
        lrparser = self.pg.build()
        lrparser_imp = LRParserImproved(lrparser.lr_table, lrparser.error_handler)
        return lrparser_imp


class LRParserImproved(rply.parser.LRParser):
    def parse(self, tokenizer, state=None):
        from rply.token import Token

        lookahead = None
        lookaheadstack = []
        processing = None

        statestack = [0]
        symstack = [Token("$end", "$end")]

        current_state = 0
        try:
            while True:
                if self.lr_table.default_reductions[current_state]:
                    t = self.lr_table.default_reductions[current_state]
                    current_state = self._reduce_production(
                        t, symstack, statestack, state
                    )
                    continue

                if lookahead is None:
                    if lookaheadstack:
                        lookahead = lookaheadstack.pop()
                    else:
                        try:
                            lookahead = next(tokenizer)
                        except StopIteration:
                            lookahead = None

                    if lookahead is None:
                        lookahead = Token(
                            "$end",
                            "$end",
                            source_pos=rply.token.SourcePosition(
                                idx=None, lineno=1, colno=1
                            ),
                        )
                    processing = lookahead

                ltype = lookahead.gettokentype()
                if ltype in self.lr_table.lr_action[current_state]:
                    t = self.lr_table.lr_action[current_state][ltype]
                    if t > 0:
                        statestack.append(t)
                        current_state = t
                        symstack.append(lookahead)
                        processing = lookahead
                        lookahead = None
                        continue
                    elif t < 0:
                        current_state = self._reduce_production(
                            t, symstack, statestack, state
                        )
                        continue
                    else:
                        n = symstack[-1]
                        return n
                else:
                    # TODO: actual error handling here
                    if self.error_handler is not None:
                        if state is None:
                            self.error_handler(lookahead)
                        else:
                            self.error_handler(state, lookahead)
                        raise AssertionError("For now, error_handler must raise.")
                    else:
                        # erase trace
                        raise rply.errors.ParsingError(
                            None, lookahead.getsourcepos()
                        ) from None
        except Exception as e:
            if isinstance(e, rply.errors.ParsingError) or isinstance(
                e, rply.errors.LexingError
            ):
                raise e
            else:
                # annotate error with position in code where it occurred
                raise rply.errors.ParsingError(
                    str(e), processing.getsourcepos()
                ) from None


def wrap_builtin_type(typ: ast.Constant, val):
    """Hmmmmmmm wraps ...."""
    if isinstance(typ, ast.PlutusData):
        assert isinstance(val, bytes), f"Expected bytes but found {type(val)}"
        return ast.data_from_cbor(val)
    if isinstance(typ, ast.BuiltinList):
        assert isinstance(val, list), f"Expected list but found {type(val)}"
        return ast.BuiltinList(
            [wrap_builtin_type(typ.sample_value, v) for v in val], typ.sample_value
        )
    if isinstance(typ, ast.BuiltinPair):
        assert isinstance(val, tuple) or isinstance(
            val, list
        ), f"Expected tuple but found {type(val)}"
        return ast.BuiltinPair(
            wrap_builtin_type(typ.l_value, val[0]),
            wrap_builtin_type(typ.r_value, val[1]),
        )
    if isinstance(typ, ast.BuiltinUnit):
        assert val is None, f"Expected () but found {type(val)}"
        return ast.BuiltinUnit()
    if isinstance(typ, ast.BuiltinByteString):
        assert isinstance(val, bytes), f"Expected bytes but found {type(val)}"
    if isinstance(typ, ast.BuiltinString):
        assert isinstance(val, str), f"Expected str but found {type(val)}"
    if isinstance(typ, ast.BuiltinInteger):
        assert isinstance(val, int), f"Expected int but found {type(val)}"
    if isinstance(typ, ast.BuiltinBool):
        assert isinstance(val, bool), f"Expected bool but found {type(val)}"
    return typ.__class__(val)
